import magent


def get_config(map_size):
    gw = magent.gridworld
    cfg = gw.Config()

    cfg.set({"map_width": map_size, "map_height": map_size})

    predator = cfg.register_agent_type(
        "predator",
        {
            'width': 2, 'length': 2, 'hp': 1, 'speed': 1,
            'view_range': gw.CircleRange(5), 'attack_range': gw.CircleRange(2),
            'attack_penalty': -0.2
        })

    prey = cfg.register_agent_type(
        "prey",
        {
            'width': 1, 'length': 1, 'hp': 1, 'speed': 1.5,
            'view_range': gw.CircleRange(4), 'attack_range': gw.CircleRange(0)
        })

    predator_group  = cfg.add_group(predator)
    prey_group = cfg.add_group(prey)

    a = gw.AgentSymbol(predator_group, index='any')
    b = gw.AgentSymbol(prey_group, index='any')

    cfg.add_reward_rule(gw.Event(a, 'attack', b), receiver=[a, b], value=[1, -1])

    return cfg



"""let saved models to play tournament"""

import os
import numpy as np
import time
import re
import math

import magent
from magent.builtin.tf_model import DeepQNetwork
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'


def play(env, handles, models, map_size, leftID, rightID, eps=0.05):
    env.reset()

    # generate map
    width = height = map_size
    init_num = map_size * map_size * 0.04
    gap = 3

    # left
    n = init_num
    side = int(math.sqrt(n)) * 2
    pos = []
    for x in range(width//2 - gap - side, width//2 - gap - side + side, 2):
        for y in range((height - side)//2, (height - side)//2 + side, 2):
            pos.append([x, y, 0])
    env.add_agents(handles[leftID], method="custom", pos=pos)

    # right
    n = init_num
    side = int(math.sqrt(n)) * 2
    pos = []
    for x in range(width//2 + gap, width//2 + gap + side, 2):
        for y in range((height - side)//2, (height - side)//2 + side, 2):
            pos.append([x, y, 0])
    env.add_agents(handles[rightID], method="custom", pos=pos)

    step_ct = 0
    done = False

    n = 2
    obs  = [[] for _ in range(n)]
    ids  = [[] for _ in range(n)]
    acts = [[] for _ in range(n)]
    nums = [env.get_num(handle) for handle in handles]

    while not done:
        # take actions for every model
        for i in range(n):
            obs[i] = env.get_observation(handles[i])
            ids[i] = env.get_agent_id(handles[i])
            models[i].infer_action(obs[i], ids[i], 'e_greedy', eps, block=False)
        for i in range(n):
            acts[i] = models[i].fetch_action()
            env.set_action(handles[i], acts[i])

        done = env.step()
        nums = [env.get_num(handle) for handle in handles]
        env.clear_dead()

        step_ct += 1
        if step_ct > 550:
            break

    return nums


def extract_model_names(savedir, name, model_class, begin=0, pick_every=4):
    if model_class is DeepQNetwork:
        prefix = 'tfdqn'
    pattern = re.compile(prefix + '_(\d*).meta')

    ret = []
    for path in os.listdir(os.path.join(savedir, name)):
        match = pattern.match(path)
        if match and int(match.group(1)) > begin:
            ret.append((savedir, name, int(match.group(1)), model_class))

    ret.sort(key=lambda x: x[2])
    ret = [ret[i] for i in range(0, len(ret), pick_every)]

    return ret


if __name__ == '__main__':
    map_size = 125
    env = magent.GridWorld("battle", map_size=map_size)
    env.set_render_dir("build/render")

    # scan file names
    model_name = extract_model_names('save_model', 'battle', DeepQNetwork, begin=0, pick_every=5)

    print("total models = %d" % len(model_name))
    print("models", [x[:-1] for x in model_name])
    handles = env.get_handles()

    def play_wrapper(model_names, n_rounds):
        time_stamp = time.time()

        models = []
        for i, item in enumerate(model_names):
            models.append(magent.ProcessingModel(env, handles[i], item[1], 0, item[-1]))

        for i, item in enumerate(model_names):
            models[i].load(item[0], item[2])

        leftID, rightID = 0, 1
        result = 0
        total_num = np.zeros(2)
        for _ in range(n_rounds):
            round_num = play(env, handles, models, map_size, leftID, rightID)
            total_num += round_num
            leftID, rightID = rightID, leftID
            result += 1 if round_num[0] > round_num[1] else 0
        result = 1.0 * result

        for model in models:
            model.quit()

        return result / n_rounds, total_num / n_rounds, time.time() - time_stamp

    detail_file = open("detail.log", "w")
    winrate_file = open("win_rate.log", "w")

    rate = [[0.0 for j in range(len(model_name))] for i in range(len(model_name))]
    for i in range(len(model_name)):
        for j in range(i+1, len(model_name)):
            rate[i][j], nums, elapsed = play_wrapper([model_name[i], model_name[j]], 6)
            rate[j][i] = 1.0 - rate[i][j]
            round_res = ("model1: %s\t model2: %s\t rate: %.2f\t num: %s\t elapsed: %.2f" %
                        (model_name[i][:-1], model_name[j][:-1], rate[i][j], list(nums), elapsed))
            print(round_res)
            detail_file.write(round_res + "\n")

        winrate_file.write("model: %s\twin rate: %.2f\n" % (model_name[i],
                                                            1.0 * sum(np.asarray(rate[i])) / (len(model_name) - 1)))

        detail_file.flush()
        winrate_file.flush()




""" some utility for call C++ code"""

from __future__ import absolute_import

import os
import ctypes
import platform
import multiprocessing


def _load_lib():
    """ Load library in build/lib. """
    cur_path = os.path.dirname(os.path.abspath(os.path.expanduser(__file__)))
    lib_path = os.path.join(cur_path, "build/")
    if platform.system() == 'Darwin':
        path_to_so_file = os.path.join(lib_path, "libmagent.dylib")
    elif platform.system() == 'Linux':
        path_to_so_file = os.path.join(lib_path, "libmagent.so")
    else:
        raise BaseException("unsupported system: " + platform.system())
    lib = ctypes.CDLL(path_to_so_file, ctypes.RTLD_GLOBAL)
    return lib


def as_float_c_array(buf):
    """numpy to ctypes array"""
    return buf.ctypes.data_as(ctypes.POINTER(ctypes.c_float))


def as_int32_c_array(buf):
    """numpy to ctypes array"""
    return buf.ctypes.data_as(ctypes.POINTER(ctypes.c_int32))


def as_bool_c_array(buf):
    """numpy to ctypes array"""
    return buf.ctypes.data_as(ctypes.POINTER(ctypes.c_bool))


if 'OMP_NUM_THREADS' not in os.environ:
    os.environ['OMP_NUM_THREADS'] = str(multiprocessing.cpu_count() // 2)
_LIB = _load_lib()



"""Deep recurrent Q network"""

import time
import os
import collections

import numpy as np
import tensorflow as tf

from .base import TFBaseModel


class DeepRecurrentQNetwork(TFBaseModel):
    def __init__(self, env, handle, name,
                 batch_size=32, unroll_step=8, reward_decay=0.99, learning_rate=1e-4,
                 train_freq=1, memory_size=20000, target_update=2000, eval_obs=None,
                 use_dueling=True, use_double=True, use_episode_train=False,
                 custom_view_space=None, custom_feature_space=None):
        """init a model

        Parameters
        ----------
        env: Environment
            environment
        handle: Handle (ctypes.c_int32)
            handle of this group, can be got by env.get_handles
        name: str
            name of this model
        learning_rate: float
        batch_size: int
        reward_decay: float
            reward_decay in TD
        train_freq: int
            mean training times of a sample
        target_update: int
            target will update every target_update batches
        memory_size: int
            weight of entropy loss in total loss
        eval_obs: numpy array
            evaluation set of observation
        use_dueling: bool
            whether use dueling q network
        use_double: bool
            whether use double q network
        custom_feature_space: tuple
            customized feature space
        custom_view_space: tuple
            customized feature space
        """
        TFBaseModel.__init__(self, env, handle, name, "tfdrqn")
        # ======================== set config  ========================
        self.env = env
        self.handle = handle
        self.view_space = custom_view_space or env.get_view_space(handle)
        self.feature_space = custom_feature_space or env.get_feature_space(handle)
        self.num_actions = env.get_action_space(handle)[0]

        self.batch_size   = batch_size
        self.unroll_step  = unroll_step
        self.handle = handle
        self.name = name
        self.learning_rate= learning_rate
        self.train_freq   = train_freq   # train time of every sample (s,a,r,s')
        self.target_update= target_update # target network update frequency
        self.eval_obs = eval_obs

        self.use_dueling  = use_dueling
        self.use_double   = use_double
        self.use_episode_train = use_episode_train
        self.skip_error   = 0
        self.pad_before_len = unroll_step - 1

        self.agent_states = {}
        self.train_ct = 0

        # ======================= build network =======================
        # input place holder
        self.target = tf.placeholder(tf.float32, [None])

        self.input_view    = tf.placeholder(tf.float32, (None,) + self.view_space, name="input_view")
        self.input_feature = tf.placeholder(tf.float32, (None,) + self.feature_space, name="input_feature")
        self.action = tf.placeholder(tf.int32, [None], name="action")
        self.mask   = tf.placeholder(tf.float32, [None], name="mask")

        self.batch_size_ph   = tf.placeholder(tf.int32, [])
        self.unroll_step_ph = tf.placeholder(tf.int32, [])

        # build graph
        with tf.variable_scope(self.name):
            with tf.variable_scope("eval_net_scope"):
                self.eval_scope_name   = tf.get_variable_scope().name
                self.qvalues, self.state_in, self.rnn_state = \
                    self._create_network(self.input_view, self.input_feature)

            with tf.variable_scope("target_net_scope"):
                self.target_scope_name = tf.get_variable_scope().name
                self.target_qvalues, self.target_state_in, self.target_rnn_state = \
                    self._create_network(self.input_view, self.input_feature)

        # loss
        self.gamma = reward_decay
        self.actions_onehot = tf.one_hot(self.action, self.num_actions)
        self.td_error = tf.square(
                self.target - tf.reduce_sum(tf.multiply(self.actions_onehot, self.qvalues), axis=1)
            )
        #self.loss = tf.reduce_mean(self.td_error)
        self.loss = tf.reduce_sum(self.td_error * self.mask) / tf.reduce_sum(self.mask)

        # train op (clip gradient)
        optimizer = tf.train.AdamOptimizer(learning_rate=learning_rate)
        gradients, variables = zip(*optimizer.compute_gradients(self.loss))
        gradients, _ = tf.clip_by_global_norm(gradients, 10.0)
        self.train_op = optimizer.apply_gradients(zip(gradients, variables))

        # target network update op
        self.update_target_op = []
        t_params = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, self.target_scope_name)
        e_params = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, self.eval_scope_name)
        for i in range(len(t_params)):
            self.update_target_op.append(tf.assign(t_params[i], e_params[i]))

        # init tensorflow session
        config = tf.ConfigProto(allow_soft_placement=True, log_device_placement=False)
        config.gpu_options.allow_growth = True
        self.sess = tf.Session(config=config)
        self.sess.run(tf.global_variables_initializer())

        # init memory buffers
        self.memory_size = memory_size
        self.replay_buffer_lens = collections.deque(maxlen=memory_size)
        self.replay_buffer = collections.deque(maxlen=memory_size)
        # item format [views, features, actions, rewards, terminals, masks, len]

        # init training buffers
        self.view_buf = np.empty((1,) + self.view_space)
        self.feature_buf = np.empty((1,) + self.feature_space)
        self.action_buf, self.reward_buf = np.empty(1, dtype=np.int32), np.empty(1)
        self.terminal_buf = np.empty(1, dtype=np.bool)

    def _create_network(self, input_view, input_feature, reuse=None):
        """define computation graph of network

        Parameters
        ----------
        input_view: tf.tensor
        input_feature: tf.tensor
            the input tensor
        """
        kernel_num  = [32, 32]
        hidden_size = [256]

        # conv
        h_conv1 = tf.layers.conv2d(input_view, filters=kernel_num[0], kernel_size=3,
                                   activation=tf.nn.relu, name="conv1", reuse=reuse)
        h_conv2 = tf.layers.conv2d(h_conv1, filters=kernel_num[1], kernel_size=3,
                                   activation=tf.nn.relu, name="conv2", reuse=reuse)
        flatten_view = tf.reshape(h_conv2, [-1, np.prod([v.value for v in h_conv2.shape[1:]])])
        h_view = tf.layers.dense(flatten_view, units=hidden_size[0], activation=tf.nn.relu,
                                 name="dense_view", reuse=reuse)

        h_emb = tf.layers.dense(input_feature,  units=hidden_size[0], activation=tf.nn.relu,
                                name="dense_emb", reuse=reuse)

        dense = tf.concat([h_view, h_emb], axis=1)

        # RNN
        state_size = hidden_size[0] * 2
        rnn_cell = tf.contrib.rnn.GRUCell(num_units=state_size)

        rnn_in = tf.reshape(dense, shape=[self.batch_size_ph, self.unroll_step_ph, state_size])
        state_in = rnn_cell.zero_state(self.batch_size_ph, tf.float32)
        rnn, rnn_state = tf.nn.dynamic_rnn(
            cell=rnn_cell, inputs=rnn_in, dtype=tf.float32, initial_state=state_in
        )
        rnn = tf.reshape(rnn, shape=[-1, state_size])

        if self.use_dueling:
            value = tf.layers.dense(dense, units=1, name="dense_value", reuse=reuse)
            advantage = tf.layers.dense(dense, units=self.num_actions, use_bias=False,
                                        name="dense_advantage", reuse=reuse)

            qvalues = value + advantage - tf.reduce_mean(advantage, axis=1, keep_dims=True)
        else:
            qvalues = tf.layers.dense(rnn, units=self.num_actions)

        self.state_size = state_size
        return qvalues, state_in, rnn_state

    def _get_agent_states(self, ids):
        """get hidden state of agents"""
        n = len(ids)
        states = np.empty([n, self.state_size])
        default = np.zeros([self.state_size])
        for i in range(n):
            states[i] = self.agent_states.get(ids[i], default)
        return states

    def _set_agent_states(self, ids, states):
        """set hidden state for agents"""
        if len(ids) <= len(self.agent_states) * 0.5:
            self.agent_states = {}
        for i in range(len(ids)):
            self.agent_states[ids[i]] = states[i]

    def infer_action(self, raw_obs, ids, policy='e_greedy', eps=0):
        """infer action for a batch of agents

        Parameters
        ----------
        raw_obs: tuple(numpy array, numpy array)
            raw observation of agents tuple(views, features)
        ids: numpy array
            ids of agents
        policy: str
            can be eps-greedy or greedy
        eps: float
            used when policy is eps-greedy

        Returns
        -------
        acts: numpy array of int32
            actions for agents
        """
        view, feature = raw_obs[0], raw_obs[1]
        n = len(ids)

        states = self._get_agent_states(ids)
        qvalues, states = self.sess.run([self.qvalues, self.rnn_state], feed_dict={
            self.input_view:    view,
            self.input_feature: feature,
            self.state_in:      states,
            self.batch_size_ph: n,
            self.unroll_step_ph: 1
        })
        self._set_agent_states(ids, states)
        best_actions = np.argmax(qvalues, axis=1)

        if policy == 'e_greedy':
            random = np.random.randint(self.num_actions, size=(n,))
            cond = np.random.uniform(0, 1, size=(n,)) < eps
            ret = np.where(cond, random, best_actions)
        elif policy == 'greedy':
            ret = best_actions

        return ret.astype(np.int32)

    def _calc_target(self, next_view, next_feature, rewards, terminal, batch_size, unroll_step):
        """calculate target value"""
        n = len(rewards)
        if self.use_double:
            t_qvalues, qvalues = self.sess.run([self.target_qvalues, self.qvalues], feed_dict={
                self.input_view:      next_view,
                self.input_feature:   next_feature,
                # self.state_in:        state_in,
                # self.target_state_in: state_in,
                self.batch_size_ph:   batch_size,
                self.unroll_step_ph:  unroll_step})
            # ignore the first value (the first value is for computing correct hidden state)
            # t_qvalues = t_qvalues.reshape([-1, unroll_step, self.num_actions])
            # t_qvalues = t_qvalues[:, 1:, :].reshape([-1, self.num_actions])
            # qvalues = qvalues.reshape([-1, unroll_step, self.num_actions])
            # qvalues = qvalues[:, 1:, :].reshape([-1, self.num_actions])
            next_value = t_qvalues[np.arange(n), np.argmax(qvalues, axis=1)]
        else:
            t_qvalues = self.sess.run(self.target_qvalues, feed_dict={
                self.input_view:      next_view,
                self.input_feature:   next_feature,
                # self.target_state_in: state_in,
                self.batch_size_ph:   batch_size,
                self.unroll_step_ph:  unroll_step})
            # t_qvalues = t_qvalues.reshape([-1, unroll_step, self.num_actions])
            # t_qvalues = t_qvalues[:,1:,:].reshape([-1, self.num_actions])
            next_value = np.max(t_qvalues, axis=1)

        target = np.where(terminal, rewards, rewards + self.gamma * next_value)

        return target

    def _add_to_replay_buffer(self, sample_buffer):
        """add samples in sample_buffer to replay buffer"""
        n = 0
        for episode in sample_buffer.episodes():
            v, f, a, r = episode.views, episode.features, episode.actions, episode.rewards

            m = len(r)

            mask = np.ones((m,))
            terminal = np.zeros((m,), dtype=np.bool)
            if episode.terminal:
                terminal[-1] = True
            else:
                mask[-1] = 0

            item = [v, f, a, r, terminal, mask, m]
            self.replay_buffer_lens.append(m)
            self.replay_buffer.append(item)

            n += m
        return n

    def train(self, sample_buffer, print_every=500):
        """ add new samples in sample_buffer to replay buffer and train
        do not keep hidden state (split episode into short sequences)

        Parameters
        ----------
        sample_buffer: magent.utility.EpisodesBuffer
            buffer contains samples
        print_every: int
            print log every print_every batches

        Returns
        -------
        loss: float
            bellman residual loss
        value: float
            estimated state value
        """
        add_num = self._add_to_replay_buffer(sample_buffer)

        batch_size = self.batch_size
        unroll_step = self.unroll_step

        # calc sample weight of episodes (i.e. their lengths)
        replay_buffer = self.replay_buffer
        replay_lens_sum = np.sum(self.replay_buffer_lens)
        weight = np.array(self.replay_buffer_lens, dtype=np.float32) / replay_lens_sum

        n_batches = self.train_freq * add_num / (batch_size * (unroll_step - self.skip_error))
        if n_batches == 0:
            return 0, 0

        max_ = batch_size * unroll_step
        batch_view = np.zeros((max_+1,) + self.view_space, dtype=np.float32)
        batch_feature = np.zeros((max_+1,) + self.feature_space, dtype=np.float32)
        batch_action = np.zeros((max_,), dtype=np.int32)
        batch_reward = np.zeros((max_,), dtype=np.float32)
        batch_terminal = np.zeros((max_,), dtype=np.bool)
        batch_mask = np.zeros((max_,), dtype=np.float32)

        # calc batch number
        n_batches = int(self.train_freq * add_num / (batch_size * (unroll_step - self.skip_error)))
        print("batches: %d  add: %d  replay_len: %d/%d" %
              (n_batches, add_num, len(self.replay_buffer), self.memory_size))

        ct = 0
        total_loss = 0
        start_time = time.time()
        # train batches
        for i in range(n_batches):
            indexes = np.random.choice(len(replay_buffer), self.batch_size, p=weight)

            batch_mask[:] = 0

            for j in range(batch_size):
                item = replay_buffer[indexes[j]]
                v, f, a, r, t = item[0], item[1], item[2], item[3], item[4]
                length = len(v)

                start = np.random.randint(length)
                real_step = min(length - start, unroll_step)

                beg = j * unroll_step
                batch_view[beg:beg+real_step]     = v[start:start+real_step]
                batch_feature[beg:beg+real_step]  = f[start:start+real_step]
                batch_action[beg:beg+real_step]   = a[start:start+real_step]
                batch_reward[beg:beg+real_step]   = r[start:start+real_step]
                batch_terminal[beg:beg+real_step] = t[start:start+real_step]
                batch_mask[beg:beg+real_step]     = 1.0

                if not t[start+real_step-1]:
                    batch_mask[beg+real_step-1] = 0

            # collect trajectories from different IDs to a single buffer
            target = self._calc_target(batch_view[1:], batch_feature[1:],
                                       batch_reward, batch_terminal, batch_size, unroll_step)

            ret = self.sess.run([self.train_op, self.loss], feed_dict={
                self.input_view:      batch_view[:-1],
                self.input_feature:   batch_feature[:-1],
                self.action:          batch_action,
                self.target:          target,
                self.mask:            batch_mask,
                self.batch_size_ph:   batch_size,
                self.unroll_step_ph:  unroll_step,
            })
            loss = ret[1]
            total_loss += loss

            if ct % self.target_update == 0:
                self.sess.run(self.update_target_op)

            if ct % print_every == 0:
                print("batch %5d, loss %.6f, qvalue %.6f" % (ct, loss, self._eval(target)))
            ct += 1
            self.train_ct += 1

        total_time = time.time() - start_time
        step_average = total_time / max(1.0, (ct / 1000.0))
        print("batches: %d,  total time: %.2f,  1k average: %.2f" % (ct, total_time, step_average))

        return total_loss / ct if ct != 0 else 0, self._eval(target)

    def train_keep_hidden(self, sample_buffer, print_every=500):
        """ add new samples in sample_buffer to replay buffer and train
            keep hidden state (split episode into small sequence, but keep hidden states)
            this means must train some episodes continuously not fully random.
            to use this training scheme, you should also modify self._calc_target

        Parameters
        ----------
        sample_buffer: magent.utility.EpisodesBuffer
            buffer contains samples
        print_every: int
            print log every print_every batches

        Returns
        -------
        loss: float
            bellman residual loss
        value: float
            estimated state value
        """

        add_num = self._add_to_replay_buffer(sample_buffer)
   
        batch_size = self.batch_size
        unroll_step = self.unroll_step

        # calc sample weight of episodes (i.e. their lengths)
        replay_buffer = self.replay_buffer
        replay_lens_sum = np.sum(self.replay_buffer_lens)
        weight = np.array(self.replay_buffer_lens, dtype=np.float32) / replay_lens_sum
   
        max_len  = self._div_round(np.max(self.replay_buffer_lens), unroll_step)
        n_batches = self.train_freq * add_num / (batch_size * unroll_step)
        if n_batches == 0:
            return 0, 0

        # allocate buffer
        max_ = batch_size * max_len
        batch_view     = np.zeros((max_+1,) + self.view_space, dtype=np.float32)
        batch_feature  = np.zeros((max_+1,) + self.feature_space, dtype=np.float32)
        batch_action   = np.zeros((max_,), dtype=np.int32)
        batch_reward   = np.zeros((max_,), dtype=np.float32)
        batch_terminal = np.zeros((max_,), dtype=np.bool)
        batch_mask     = np.zeros((max_,), dtype=np.float32)
        batch_hidden   = np.zeros((batch_size, self.state_size), dtype=np.float32)
        batch_pick     = np.zeros((batch_size, max_len), dtype=np.bool)
        pick_buffer    = np.arange(max_, dtype=np.int32)

        print("batches: %d  add: %d  replay_len: %d, %d/%d" %
              (n_batches, add_num, replay_lens_sum, len(self.replay_buffer), self.memory_size))

        start_time = time.time()
        total_loss = 0
        ct = 0
        while ct < n_batches:
            # random sample agent episodes (sequence)
            indexs = np.random.choice(len(replay_buffer), self.batch_size, p=weight)
            train_length = 0
            to_sort = []

            # collect length and sort
            for j, index in enumerate(indexs):
                length = replay_buffer[index][-1]
                length = self._div_round(length, unroll_step)
                train_length = max(train_length, length)
                to_sort.append([index, length])
            to_sort.sort(key=lambda x: -x[1])

            # merge short episodes to long episodes (use greedy method)
            merged = [False for _ in range(batch_size)]
            rows = []
            for j in range(len(to_sort)):
                if merged[j]:
                    continue
                row = [to_sort[j][0]]
                now_len = to_sort[j][1]
                if True: # use compress
                    for k in range(j+1, batch_size):
                        if now_len + to_sort[k][1] <= train_length:
                            row.append(to_sort[k][0])
                            now_len += to_sort[k][1]
                            merged[k] = True
                    rows.append(row)
            n_rows = len(rows)

            batch_reset = np.zeros((train_length, batch_size), dtype=np.bool)
            batch_mask[:] = 0

            # copy from replay buffer to batch buffer
            for j, row in enumerate(rows):
                beg = j * max_len
                init_beg = beg
                # fill a row
                for index in row:
                    v, f, a, r, terminal, mask, x = replay_buffer[index]

                    batch_reset[(beg - init_beg)/unroll_step, j] = True
                    batch_view[beg:beg+x] = v
                    batch_feature[beg:beg+x] = f
                    batch_action[beg:beg+x] = a
                    batch_reward[beg:beg+x] = r
                    batch_terminal[beg:beg+x] = terminal
                    batch_mask[beg:beg+x] = mask

                    beg += self._div_round(x, unroll_step)

            # train steps
            for j in range((train_length + unroll_step - 1) / unroll_step):
                batch_pick[:] = False
                batch_pick[:n_rows, j * unroll_step:(j+1) * unroll_step] = True

                pick = pick_buffer[batch_pick.reshape(-1)].reshape(n_rows, unroll_step)
                next_pick = np.empty((n_rows, unroll_step + 1), dtype=np.int32)  # next pick choose one more state than pick
                next_pick[:, :unroll_step] = pick
                next_pick[:, unroll_step]  = pick[:, -1] + 1
                pick = pick.reshape(-1)
                next_pick = next_pick.reshape(-1)

                steps = len(pick) / n_rows
                assert steps > 0
                if np.sum(batch_mask[pick]) < 1:
                    continue

                batch_hidden[batch_reset[j]] = np.zeros_like(batch_hidden[0])

                batch_target = self._calc_target(batch_view[next_pick], batch_feature[next_pick],
                                                 batch_reward[pick], batch_terminal[pick], batch_hidden[:n_rows],
                                                 n_rows, steps + 1)
                ret = self.sess.run(
                    [self.train_op, self.loss, self.rnn_state],
                    feed_dict={
                        self.input_view:      batch_view[pick],
                        self.input_feature:   batch_feature[pick],
                        self.action:          batch_action[pick],
                        self.target:          batch_target,
                        self.mask:            batch_mask[pick],
                        self.state_in:        batch_hidden[:n_rows],
                        self.batch_size_ph:   n_rows,
                        self.unroll_step_ph:  steps
                })
                loss, batch_hidden[:n_rows] = ret[1], ret[2]
                total_loss += loss

                if ct % self.target_update == 0:
                    self.sess.run(self.update_target_op)

                if ct % print_every == 0:
                    print("batches %5d, mask %d/%d (%d), loss %.6f, qvalue %.6f" %
                          (ct, sum(batch_mask), n_rows * train_length, n_rows, loss, self._eval(batch_target)))
                ct += 1
                self.train_ct += 1

        total_time = time.time() - start_time
        step_average = total_time / max(1.0, (ct / 1000.0))
        print("batches: %d,  total time: %.2f,  1k average: %.2f" % (ct, total_time, step_average))

        return round(total_loss / ct if ct != 0 else 0, 6), self._eval(batch_target)

    @staticmethod
    def _div_round(x, divisor):
        """round up to nearest integer that are divisible by divisor"""
        return (x + divisor - 1) / divisor * divisor

    def _eval(self, target):
        """evaluate estimated q value"""
        if self.eval_obs is None:
            return np.mean(target)
        else:
            return np.mean(self.sess.run(self.target_qvalues, feed_dict={
                self.input_view:      self.eval_obs[0],
                self.input_feature:   self.eval_obs[1],
                self.batch_size_ph:   self.eval_obs[0].shape[0],
                self.unroll_step_ph: 1
            }))

    def get_info(self):
        """get information of model"""
        return "tfdrqn train_time: %d" % (self.train_ct)



"""plot general log file according to given indexes"""

import sys
import matplotlib.pyplot as plt
import numpy as np

filename = sys.argv[1]

data = []

with open(filename, 'r') as fin:
    for line in fin.readlines():
        items = line.split('\t')

        row = []
        for item in items[1:]:
            t = eval(item.split(':')[1])
            if isinstance(t, list):
                for x in t:
                    row.append(x)
            else:
                row.append(t)
        if len(row) > 0:
            data.append(row)

data = np.array(data)


for index in sys.argv[2:]:
    index = int(index)
    plt.plot(data[:, index])
plt.show()



import random

from .base_server import BaseServer


class RandomServer(BaseServer):
    def __init__(self, agent_number=1000, group_number=20, map_size=100, shape_range=3, speed=5, event_range=100):
        self._data = {}
        self._map_size = map_size
        self._number = agent_number
        for i in range(agent_number):
            self._data.setdefault(i, [
                random.randint(0, map_size - 1),
                random.randint(0, map_size - 1),
                random.randint(0, group_number - 1)
            ])
        self._group = []
        for i in range(group_number):
            self._group.append([
                random.randint(1, shape_range),
                random.randint(1, shape_range),
                random.randint(0, 255),
                random.randint(0, 255),
                random.randint(0, 255)
            ])
        self._speed = speed
        self._event_range = event_range
        self._map_size = map_size

    def get_group_info(self):
        return self._group

    def get_static_info(self):
        return {"wall": []}

    def get_data(self, frame_id, x_range, y_range):
        result = {}
        event = []
        for i in self._data:
            olddata = self._data[i]
            data = [0, 0, 0]
            data[0] = olddata[0] + random.randint(-self._speed, self._speed)
            data[1] = olddata[1] + random.randint(-self._speed, self._speed)
            data[0] = min(max(data[0], 0), self._map_size - 1)
            data[1] = min(max(data[1], 0), self._map_size - 1)
            data[2] = olddata[2]
            self._data[i] = data
            if (x_range[0] <= data[0] <= x_range[1] and y_range[0] <= data[1] <= y_range[1]) or \
                    (x_range[0] <= olddata[0] <= x_range[1] and y_range[0] <= olddata[1] <= y_range[1]):
                result.setdefault(i, olddata)
        event_number = random.randint(0, self._event_range)
        for i in range(event_number):
            agent_id, _ = random.choice(self._data.items())
            event.append(
                (
                    agent_id,
                    random.randint(0, self._map_size - 1),
                    random.randint(0, self._map_size - 1)
                )
            )
        return result, event

    def add_agents(self, x, y, g):
        self._data.setdefault(self._number, (x, y, g))
        self._number += 1
        
    def get_map_size(self):
        return self._map_size, self._map_size




"""test examples"""

import os
import time

source = [
    "examples/train_tiger.py",
    "examples/train_pursuit.py",
    "examples/train_gather.py",
    "examples/train_battle.py",
    "examples/train_single.py",
    "examples/train_arrange.py",
    "examples/train_multi.py",
]


def do_cmd(cmd):
    tic = time.time()
    print(cmd)
    assert os.system(cmd) == 0
    return time.time() - tic


start = time.time()
for item in source:
    run_cmd = "python %s --train --n_round 1" % item
    do_cmd(run_cmd)

    change_cmd = "sed -i 's/tf_model/mx_model/g' %s" % item
    do_cmd(change_cmd)

    do_cmd(run_cmd)

print("test examples done", time.time() - start)



"""advantage actor critic"""

import os
import time

import numpy as np
import mxnet as mx

from .base import MXBaseModel


class AdvantageActorCritic(MXBaseModel):
    def __init__(self, env, handle, name, eval_obs=None,
                 batch_size=64, reward_decay=0.99, learning_rate=1e-3,
                 train_freq=1, value_coef=0.1, ent_coef=0.1,
                 custom_view_space=None, custom_feature_space=None,
                 *args, **kwargs):
        """init a model

        Parameters
        ----------
        env: Environment
            environment
        handle: Handle (ctypes.c_int32)
            handle of this group, can be got by env.get_handles
        name: str
            name of this model
        learning_rate: float
        batch_size: int
        reward_decay: float
            reward_decay in TD
        eval_obs: numpy array
            evaluation set of observation
        train_freq: int
            mean training times of a sample
        ent_coef: float
            weight of entropy loss in total loss
        value_coef: float
            weight of value loss in total loss
        custom_feature_space: tuple
            customized feature space
        custom_view_space: tuple
            customized feature space
        """
        MXBaseModel.__init__(self, env, handle, name, "mxa2c")
        # ======================== set config  ========================
        self.env = env
        self.handle = handle
        self.view_space = custom_view_space or env.get_view_space(handle)
        self.feature_space = custom_feature_space or env.get_feature_space(handle)
        self.num_actions  = env.get_action_space(handle)[0]
        self.reward_decay = reward_decay

        self.batch_size   = batch_size
        self.learning_rate= learning_rate
        self.train_freq   = train_freq   # train time of every sample (s,a,r,s')
        self.eval_obs     = eval_obs

        self.value_coef = value_coef
        self.ent_coef   = ent_coef

        self.train_ct = 0

        # ======================= build network =======================
        self.ctx = self._get_ctx()

        self.input_view = mx.sym.var("input_view")
        self.input_feature = mx.sym.var("input_feature")

        policy, value = self._create_network(self.input_view, self.input_feature)

        log_policy = mx.sym.log(policy)
        out_policy = mx.sym.BlockGrad(policy)

        neg_entropy = ent_coef * mx.sym.sum(policy * log_policy, axis=1)
        neg_entropy = mx.sym.MakeLoss(data=neg_entropy)

        self.sym = mx.sym.Group([log_policy, value, neg_entropy, out_policy])
        self.model = mx.mod.Module(self.sym, data_names=['input_view', 'input_feature'],
                                   label_names=None, context=self.ctx)

        # bind (set initial batch size)
        self.bind_size = batch_size
        self.model.bind(data_shapes=[('input_view', (batch_size,) + self.view_space),
                                     ('input_feature', (batch_size,) + self.feature_space)],
                        label_shapes=None)

        # init params
        self.model.init_params(initializer=mx.init.Xavier())
        self.model.init_optimizer(optimizer='adam', optimizer_params={
            'learning_rate': learning_rate,
            'clip_gradient': 10,
        })

        # init training buffers
        self.view_buf = np.empty((1,) + self.view_space)
        self.feature_buf = np.empty((1,) + self.feature_space)
        self.action_buf = np.empty(1, dtype=np.int32)
        self.advantage_buf, self.value_buf = np.empty(1), np.empty(1)
        self.terminal_buf = np.empty(1, dtype=np.bool)

        # print("parameters", self.model.get_params())
        # mx.viz.plot_network(self.output).view()

    def _create_network(self, input_view, input_feature):
        """define computation graph of network

        Parameters
        ----------
        view_space: tuple
        feature_space: tuple
            the input shape
        """
        kernel_num = [32, 32]
        hidden_size = [256]

        if False:
            h_conv1 = mx.sym.Convolution(data=input_view, kernel=(3, 3),
                                         num_filter=kernel_num[0], layout="NHWC")
            h_conv1 = mx.sym.Activation(data=h_conv1, act_type="relu")
            h_conv2 = mx.sym.Convolution(data=h_conv1, kernel=(3, 3),
                                         num_filter=kernel_num[1], layout="NHWC")
            h_conv2 = mx.sym.Activation(data=h_conv2, act_type="relu")
        else:
            input_view = mx.sym.flatten(data=input_view)
            h_conv2 = mx.sym.FullyConnected(input_view, num_hidden=hidden_size[0])
            h_conv2 = mx.sym.Activation(data=h_conv2, act_type="relu")

        flatten_view = mx.sym.flatten(data=h_conv2)
        h_view = mx.sym.FullyConnected(data=flatten_view, num_hidden=hidden_size[0])
        h_view = mx.sym.Activation(data=h_view, act_type="relu")

        h_emb = mx.sym.FullyConnected(data=input_feature, num_hidden=hidden_size[0])
        h_emb = mx.sym.Activation(data=h_emb, act_type="relu")

        dense = h_view + h_emb

        policy = mx.sym.FullyConnected(data=dense, num_hidden=self.num_actions, no_bias=True)
        policy = mx.sym.SoftmaxActivation(data=policy)
        policy = mx.sym.clip(data=policy, a_min=1e-5, a_max=1 - 1e-5)
        value  = mx.sym.FullyConnected(data=dense, num_hidden=1)

        return policy, value

    def infer_action(self, raw_obs, ids, policy="e_greedy", eps=0):
        """infer action for a batch of agents

        Parameters
        ----------
        raw_obs: tuple(numpy array, numpy array)
            raw observation of agents tuple(views, features)
        ids: numpy array
            ids of agents

        Returns
        -------
        acts: numpy array of int32
            actions for agents
        """
        view, feature = raw_obs[0], raw_obs[1]
        n = len(view)

        ret = np.empty(n, dtype=np.int32)
        self._reset_bind_size(n)
        data_batch = mx.io.DataBatch(data=[mx.nd.array(view), mx.nd.array(feature)])
        self.model.forward(data_batch, is_train=False)
        policy = self.model.get_outputs()[3].asnumpy()

        actions = np.arange(self.num_actions)
        for i in range(n):
            ret[i] = np.random.choice(actions, p=policy[i])
        return ret

    def train(self, sample_buffer, print_every=1000):
        """feed new data sample and train

        Parameters
        ----------
        sample_buffer: magent.utility.EpisodesBuffer
            buffer contains samples

        Returns
        -------
        loss: list
            policy gradient loss, critic loss, entropy loss
        value: float
            estimated state value
        """
        # calc buffer size
        n = 0
        for episode in sample_buffer.episodes():
            if episode.terminal:
                n += len(episode.rewards)
            else:
                n += len(episode.rewards) - 1

        if n == 0:
            return [0.0, 0.0, 0.0], 0.0

        # resize to the new size
        self.view_buf.resize((n,) + self.view_space)
        self.feature_buf.resize((n,) + self.feature_space)
        self.action_buf.resize(n)
        self.value_buf.resize(n)
        self.advantage_buf.resize(n)
        view, feature = self.view_buf, self.feature_buf
        action, value = self.action_buf, self.value_buf
        advantage     = self.advantage_buf

        ct = 0
        gamma = self.reward_decay
        # collect episodes from multiple separate buffers to a continuous buffer

        for episode in sample_buffer.episodes():
            v, f, a, r = episode.views, episode.features, episode.actions, episode.rewards

            m = len(episode.rewards)
            self._reset_bind_size(m)
            data_batch = mx.io.DataBatch(data=[mx.nd.array(v), mx.nd.array(f)])
            self.model.forward(data_batch, is_train=False)
            value = self.model.get_outputs()[1].asnumpy().flatten()

            delta_t = np.empty(m)
            if episode.terminal:
                delta_t[:m-1] = r[:m-1] + gamma * value[1:m] - value[:m-1]
                delta_t[m-1]  = r[m-1]  + gamma * 0          - value[m-1]
            else:
                delta_t[:m-1] = r[:m-1] + gamma * value[1:m] - value[:m-1]
                m -= 1
                v, f, a = v[:-1], f[:-1], a[:-1]

            if m == 0:
                continue

            # discount advantage
            keep = 0
            for i in reversed(range(m)):
                keep = keep * gamma + delta_t[i]
                advantage[ct+i] = keep

            view[ct:ct+m] = v
            feature[ct:ct+m] = f
            action[ct:ct+m]  = a
            ct += m
        assert n == ct

        n = len(advantage)
        neg_advantage = -advantage

        neg_advs_np = np.zeros((n, self.num_actions), dtype=np.float32)
        neg_advs_np[np.arange(n), action] = neg_advantage
        neg_advs = mx.nd.array(neg_advs_np)

        # the grads of values are exactly negative advantages
        v_grads = mx.nd.array(self.value_coef * (neg_advantage[:, np.newaxis]))
        data_batch = mx.io.DataBatch(data=[mx.nd.array(view), mx.nd.array(feature)])
        self._reset_bind_size(n)
        self.model.forward(data_batch, is_train=True)
        self.model.backward(out_grads=[neg_advs, v_grads])
        self.model.update()
        log_policy, value, entropy_loss, _ = self.model.get_outputs()

        value = mx.nd.mean(value).asnumpy()[0]
        log_policy = log_policy.asnumpy()[np.arange(n), action]
        pg_loss = np.mean(neg_advantage * log_policy)
        entropy_loss = np.mean(entropy_loss.asnumpy())
        value_loss = self.value_coef * np.mean(np.square(advantage))

        print("sample %d  %.4f %.4f %.4f %.4f" % (n, pg_loss, value_loss, entropy_loss, value))
        return [pg_loss, value_loss, entropy_loss], value

    def _reset_bind_size(self, new_size):
        """reset input shape of the model

        Parameters
        ----------
        new_size: int
            new batch size
        """
        if self.bind_size == new_size:
            return
        else:
            self.bind_size = new_size
            self.model.reshape(
                data_shapes=[
                    ('input_view',    (new_size,) + self.view_space),
                    ('input_feature', (new_size,) + self.feature_space)],
            )

    def get_info(self):
        """get information of the model

        Returns
        -------
        info: string
        """
        return "mx dqn train_time: %d" % (self.train_ct)



"""do search task"""

import os
import sys
import argparse
import time

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'


def do_task(task_item):
    recorder = open(task_item["name"] + "-rec.out", "w")
    for value in task_item["arg_value"]:
        tmp_name = task_item["name"] + "-" + value
        cmd = " ".join([task_item["prefix"], task_item["arg_name"], value])
        # cmd = "time " + cmd
        cmd += " --name " + tmp_name
        cmd = cmd + " >> " + tmp_name + ".out"
        print("%s : %s" % (tmp_name, cmd))
        start = time.time()
        os.system(cmd)
        use_time = time.time() - start
        recorder.write("log_file: %s\t time: %.2f\n" % (tmp_name + ".log", use_time))
    recorder.close()




"""A random agent"""

import numpy as np

from magent.model import BaseModel


class RandomActor(BaseModel):
    def __init__(self, env, handle, *args, **kwargs):
        BaseModel.__init__(self, env, handle)

        self.env = env
        self.handle = handle
        self.n_action = env.get_action_space(handle)[0]

    def infer_action(self, obs, *args, **kwargs):
        num = len(obs[0])
        actions = np.random.randint(self.n_action, size=num, dtype=np.int32)
        return actions



import time

import numpy as np
import random
import magent
from magent.builtin.tf_model import DeepQNetwork
from magent.renderer.server import BaseServer
from magent.utility import FontProvider


def remove_wall(d, cur_pos, wall_set, unit):
    if d == 0:
        for i in range(0, unit):
            for j in range(0, unit):
                temp = (cur_pos[0] + i, cur_pos[1] + unit + j)
                if temp in wall_set:
                    wall_set.remove(temp)
    elif d == 1:
        for i in range(0, unit):
            for j in range(0, unit):
                temp = (cur_pos[0] - unit + i, cur_pos[1] + j)
                if temp in wall_set:
                    wall_set.remove(temp)
    elif d == 2:
        for i in range(0, unit):
            for j in range(0, unit):
                temp = (cur_pos[0] + i, cur_pos[1] - unit + j)
                if temp in wall_set:
                    wall_set.remove(temp)
    elif d == 3:
        for i in range(0, unit):
            for j in range(0, unit):
                temp = (cur_pos[0] + unit + i, cur_pos[1] + j)
                if temp in wall_set:
                    wall_set.remove(temp)


def dfs(x, y, width, height, unit, wall_set):
    pos = set()
    trace = list()
    pos.add((x, y))
    trace.append((x, y))

    max_x = x + width
    max_y = y + height

    d = random.choice(range(4))
    pos_list = []
    flag = 0
    while len(trace) > 0:
        if flag == 4:
            cur_pos = trace[-1]
            trace.pop()
            if random.choice(range(2)) == 0:
                remove_wall(d, cur_pos, wall_set, unit)
            flag = 0
        if len(trace) == 0:
            break
        cur_pos = list(trace[-1])
        if d == 0:
            cur_pos[1] = max(y, cur_pos[1] - 2 * unit)
        elif d == 1:
            cur_pos[0] = min(max_x, cur_pos[0] + 2 * unit)
        elif d == 2:
            cur_pos[1] = min(max_y, cur_pos[1] + 2 * unit)
        elif d == 3:
            cur_pos[0] = max(x, cur_pos[0] - 2 * unit)
        if tuple(cur_pos) in pos:
            d = (d + 1) % 4
            flag += 1
        else:
            remove_wall(d, cur_pos, wall_set, unit)
            trace.append(tuple(cur_pos))
            pos.add(tuple(cur_pos))
            d = random.choice(range(4))


def clean_pos_set_convert_to_list(pos_set, pos_list):
    for v in pos_list:
        if v in pos_set:
            pos_set.remove(v)
    return list(pos_set)


def draw_line(x, y, width, height):
    pos_set = []
    for r in range(height):
        for c in range(width):
            pos_set.append((x + c, y + r))
    return pos_set


def open_the_door(x_s, y_s, w, h, unit):
    pos_list = []
    n_door = 15
    random_horizon_list_x = [x_s + (2 * np.random.choice(w // 2 // unit, n_door) + 1) * unit, x_s + (2 * np.random.choice(w // 2 // unit, n_door) - 1) * unit]
    random_vertical_list_y = [y_s + (2 * np.random.choice(h // 2 // unit, n_door) + 1) * unit, y_s + (2 * np.random.choice(h // 2 // unit, n_door) + 1) * unit]

    y_e = y_s + h - unit
    for v in random_horizon_list_x[0]:
        pos_list.extend([(v, y_s), (v + 1, y_s), (v, y_s + 1), (v + 1, y_s + 1)])
    for v in random_horizon_list_x[1]:
        pos_list.extend([(v, y_e), (v + 1, y_e), (v, y_e + 1), (v + 1, y_e + 1)])

    x_e = x_s + w - unit
    for v in random_vertical_list_y[0]:
        pos_list.extend([(x_s, v), (x_s, v + 1), (x_s + 1, v), (x_s + 1, v + 1)])
    for v in random_vertical_list_y[1]:
        pos_list.extend([(x_e, v), (x_e, v + 1), (x_e + 1, v), (x_e + 1, v + 1)])

    return pos_list


def create_maze(pos, width, height, unit, font_area):
    # draw block: with rect: left(x), top(y), width, height
    pos_set = []
    for i in range(height):
        if i % 2 == 0:
            pos_set.extend(draw_line(pos[0], pos[1] + i * unit, width * unit, unit))
            pos_set.extend(draw_line(pos[0], pos[1] + font_area[1] + i * unit, width * unit, unit))
            pos_set.extend(draw_line(pos[0] + i * unit, pos[1] + height * unit, unit, font_area[1]))
            pos_set.extend(draw_line(pos[0] + font_area[0] + i * unit, pos[1] + height * unit, unit, font_area[1]))

    for i in range(width):
        if i % 2 == 0:
            pos_set.extend(draw_line(pos[0] + i * unit, pos[1], unit, height * unit))
            pos_set.extend(draw_line(pos[0] + i * unit, pos[1] + font_area[1], unit, height * unit))
            pos_set.extend(draw_line(pos[0], pos[1] + i * unit, height * unit, unit))
            pos_set.extend(draw_line(pos[0] + font_area[0], pos[1] + i * unit, height * unit, unit))

    pos_set = set(pos_set)

    dfs(pos[0] + 2, pos[1] + 2, (width - 1) * unit, (height - 1) * unit, unit, pos_set)  # north
    dfs(pos[0] + 2, pos[1] + (height - 2) * unit, (height - 1) * unit, (width + 3) * unit, unit, pos_set)  # west
    dfs(pos[0] + height * unit, pos[1] + font_area[1] - unit, (width - height) * unit, (height - 1) * unit, unit, pos_set)  # south
    dfs(pos[0] + font_area[0] - unit, pos[1] + (height - 2) * unit, (height - 1) * unit, font_area[1] - (height + 1) * unit, unit, pos_set)  # east

    temp = []
    temp.extend(open_the_door(pos[0], pos[1], font_area[0] + height * unit, font_area[1] + height * unit, unit))
    res = clean_pos_set_convert_to_list(pos_set, temp)
    return res


def load_config(map_size):
    gw = magent.gridworld
    cfg = gw.Config()

    cfg.set({"map_width": map_size, "map_height": map_size})
    cfg.set({"minimap_mode": True})
    cfg.set({"embedding_size": 12})

    goal = cfg.register_agent_type(
        "goal",
        {'width': 1, 'length': 1,

         'can_absorb': True
         }
    )

    agent = cfg.register_agent_type(
        "agent",
        {'width': 1, 'length': 1, 'hp': 10, 'speed': 2,
         'view_range': gw.CircleRange(6),
         'damage': 2, 'step_recover': -10.0/400,

         'step_reward': 0,
         })

    g_goal = cfg.add_group(goal)
    g_agent = cfg.add_group(agent)

    g = gw.AgentSymbol(g_goal, 'any')
    a = gw.AgentSymbol(g_agent, 'any')

    cfg.add_reward_rule(gw.Event(a, 'collide', g), receiver=a, value=10)

    return cfg


def generate_map(mode, env, map_size, goal_handle, handles, messages, font):
    # pre-process message
    max_len = 8
    new = []
    for msg in messages:
        if len(msg) > max_len:
            for i in range(0, len(msg), max_len):
                new.append(msg[i:i+max_len])
        else:
            new.append(msg)
    messages = new

    center_x, center_y = map_size // 2, map_size // 2

    # create maze
    if mode == 1:
        radius = 90
        pos_list = create_maze([center_x - radius, center_y - radius], radius + 1, 15, 2, font_area=[radius * 2 - 28, radius * 2 - 28])
        env.add_walls(method="custom", pos=pos_list)

    def add_square(pos, side, gap):
        side = int(side)
        for x in range(center_x - side//2, center_x + side//2 + 1, gap):
            pos.append([x, center_y - side//2])
            pos.append([x, center_y + side//2])
        for y in range(center_y - side//2, center_y + side//2 + 1, gap):
            pos.append([center_x - side//2, y])
            pos.append([center_x + side//2, y])

    def draw(base_x, base_y, scale, data):
        w, h = len(data), len(data[0])
        pos = []
        for i in range(w):
            for j in range(h):
                if data[i][j] == 1:
                    start_x = i * scale + base_y
                    start_y = j * scale + base_x
                    for x in range(start_x, start_x + scale):
                        for y in range(start_y, start_y + scale):
                            pos.append([y, x])

        env.add_agents(goal_handle, method="custom", pos=pos)

    base_y = (map_size - len(messages) * font.height) // 2
    for message in messages:
        base_x = (map_size - len(message) * font.width) // 2
        scale = 1
        for x in message:
            data = font.get(x)
            draw(base_x, base_y, scale, data)
            base_x += font.width
        base_y += font.height + 1

    alpha_goal_num = env.get_num(goal_handle)

    # agent
    pos = []

    add_square(pos, map_size * 0.95, 1)
    add_square(pos, map_size * 0.90, 1)
    add_square(pos, map_size * 0.85, 1)
    add_square(pos, map_size * 0.80, 1)

    pos = np.array(pos)
    pos = pos[np.random.choice(np.arange(len(pos)), int(alpha_goal_num * 1.6), replace=False)]

    env.add_agents(handles[0], method="custom", pos=pos)


class ArrangeServer(BaseServer):
    def get_banners(self, frame_id, resolution):
        return []

    def keydown(self, frame_id, key, mouse_x, mouse_y):
        return False

    def get_status(self, frame_id):
        if self.done:
            return None
        else:
            return True

    def get_endscreen(self, frame_id):
        return []

    def mousedown(self, frame_id, key, mouse_x, mouse_y):
        return False

    def get_info(self):
        ret = self.env._get_groups_info()
        ret[1] = ret[0]
        return (self.map_size, self.map_size), ret, {'wall': self.env._get_walls_info()}

    def __init__(self, path="data/arrange_model", messages=None, mode=1):
        # some parameter
        map_size = 250
        eps = 0.15

        # init the game
        env = magent.GridWorld(load_config(map_size))
        font = FontProvider('data/font_8x8/basic.txt')

        handles = env.get_handles()
        food_handle, handles = handles[0], handles[1:]
        models = []
        models.append(DeepQNetwork(env, handles[0], 'arrange', use_conv=True))

        # load model
        models[0].load(path, 10)

        # init environment
        env.reset()
        generate_map(mode, env, map_size, food_handle, handles, messages, font)

        # save to member variable
        self.env = env
        self.food_handle = food_handle
        self.handles = handles
        self.eps = eps
        self.models = models
        self.done = False
        self.map_size = map_size
        self.new_rule_ct = 0
        self.pos_reward_ct = set()
        self.num = None

        self.ct = 0

    def step(self):
        handles = self.handles
        models = self.models
        env = self.env

        center_x = self.map_size // 2
        center_y = self.map_size

        for j in range(2):
            obs = [env.get_observation(handle) for handle in handles]
            ids = [env.get_agent_id(handle) for handle in handles]

            for i in range(len(handles)):
                if self.new_rule_ct > 0:
                    obs[i][1][:, 10:12] = 0
                else:
                    obs[i][1][:, 10:12] = 1
                acts = models[i].infer_action(obs[i], ids[i], 'e_greedy', eps=self.eps)
                env.set_action(handles[i], acts)

            done = env.step()

            goal_num = env.get_num(self.food_handle)
            rewards = env.get_reward(handles[0])

            for id_, r in zip(ids[0], rewards):
                if r > 0.05 and id_ not in self.pos_reward_ct:
                    self.pos_reward_ct.add(id_)

            if 1.0 * len(self.pos_reward_ct) / goal_num >= 0.99:
                self.new_rule_ct += 1
            
            self.num = [env.get_num(handle) for handle in [self.food_handle] + handles]
            env.clear_dead()

            if done:
                break

        return done

    def get_data(self, frame_id, x_range, y_range):
        start = time.time()

        if not self.done:
            self.done = self.step()
        print(self.done)
        if self.done:
            print("done!")

        pos, event = self.env._get_render_info(x_range, y_range)
        print(" fps ", 1 / (time.time() - start))
        return pos, event

    def add_agents(self, x, y, g):
        pos = []
        for i in range(-3, 3):
            for j in range(-3, 3):
                pos.append((x + i, y + j))
        self.env.add_agents(self.handles[g], method="custom", pos=pos)
        
    def get_map_size(self):
        return self.map_size, self.map_size

    def get_numbers(self):
        return self.num



from abc import ABCMeta, abstractmethod


class BaseRenderer:
    __metaclass__ = ABCMeta

    def __init__(self):
        pass

    @abstractmethod
    def start(self, *args, **kwargs):
        pass



"""test fps"""

import os
import sys
import magent
import argparse


if len(sys.argv) < 2:
    print("usage python test_fps.py max_gpu frame")

parser = argparse.ArgumentParser()
parser.add_argument("--max_gpu", type=int, default=0)
parser.add_argument("--frame", type=str, default='tf')
parser.add_argument("--name", type=str, default="fps")
args = parser.parse_args()

tmp_name = 'tmp-' + args.name
max_gpu = args.max_gpu
framework = args.frame

number = [1000, 10000, 100000, 1000000]
gpus   = range(max_gpu+1)

ret = []

for n in number:
    row = []
    for g in gpus:
        n_step = 30000000 / n
        cmd = ("python scripts/test/test_1m.py --n_step %d --agent_number %d --num_gpu %d --frame %s > /dev/shm/aha "
               "&& cat /dev/shm/aha | grep FPS > %s" % (n_step, n, g, framework, tmp_name))
        if n < 1000000:
            cmd = 'OMP_NUM_THREADS=8  ' + cmd
        else:
            cmd = 'OMP_NUM_THREADS=16 ' + cmd
        print(cmd)
        os.system(cmd)
        with open(tmp_name) as fin:
            line = fin.readline()
            x = eval(line)[1]
        row.append(x)
        print(x)

    ret.append(row)

for row in ret:
    print(magent.round(row))



"""test one million random agents"""

import time

import magent
import os
import math
import argparse
from magent.builtin.rule_model import RandomActor

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

def load_forest(map_size):
    gw = magent.gridworld
    cfg = gw.Config()

    cfg.set({"map_width": map_size, "map_height": map_size})

    predator = cfg.register_agent_type(
        "predator",
        {
            'width': 2, 'length': 2, 'hp': 1, 'speed': 1,
            'view_range': gw.CircleRange(5), 'attack_range': gw.CircleRange(2),
            'attack_penalty': -0.2
        })

    prey = cfg.register_agent_type(
        "prey",
        {
            'width': 1, 'length': 1, 'hp': 1, 'speed': 1.5,
            'view_range': gw.CircleRange(4), 'attack_range': gw.CircleRange(0)
        })

    predator_group  = cfg.add_group(predator)
    prey_group = cfg.add_group(prey)

    a = gw.AgentSymbol(predator_group, index='any')
    b = gw.AgentSymbol(prey_group, index='any')

    cfg.add_reward_rule(gw.Event(a, 'attack', b), receiver=[a, b], value=[1, -1])

    return cfg


def measure_time(msg, func, *args, **kwargs):
    start_time = time.time()
    ret = func(*args, **kwargs)
    print("%-11s %.5f" % (msg, time.time() - start_time))
    return ret


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--n_step", type=int, default=20)
    parser.add_argument("--agent_number", type=int, default=1000)
    parser.add_argument("--num_gpu", type=int, default=0)
    parser.add_argument('--frame', default='tf', choices=['tf', 'mx'])
    args = parser.parse_args()

    n_step = args.n_step
    agent_number = args.agent_number
    skip = 20    # warm up steps
    n_step += skip

    # init the game "forest" (or "battle" here)
    env = magent.GridWorld(load_forest(int(math.sqrt(agent_number * 20))))
    env.reset()

    # add two groups of animals
    deer_handle, tiger_handle = env.get_handles()

    env.add_walls(method="random", n=agent_number / 10)
    env.add_agents(deer_handle,  method="random", n=agent_number / 2)
    env.add_agents(tiger_handle, method="random", n=agent_number / 2)

    # init two models
    if args.num_gpu == 0:
        model1 = RandomActor(env, deer_handle, "deer")
        model2 = RandomActor(env, tiger_handle, "tiger")
    else:
        if args.frame == 'tf':
            from magent.builtin.tf_model import DeepQNetwork
        else:
            from magent.builtin.mx_model import DeepQNetwork
        model1 = DeepQNetwork(env, deer_handle, "deer", num_gpu=args.num_gpu, infer_batch_size=100000)
        model2 = DeepQNetwork(env, tiger_handle, "tiger", num_gpu=args.num_gpu, infer_batch_size=100000)

    total_reward = 0

    print(env.get_view_space(deer_handle))
    print(env.get_view_space(tiger_handle))

    total_time = 0

    for i in range(n_step):
        print("===== step %d =====" % i)
        start_time = time.time()

        obs_1 = measure_time("get obs 1", env.get_observation, deer_handle)
        acts_1 = measure_time("infer act 1", model1.infer_action, obs_1, None)
        measure_time("set act 1", env.set_action, deer_handle, acts_1)

        obs_2 = measure_time("get obs 2", env.get_observation, tiger_handle)
        acts_2 = measure_time("infer act 2", model2.infer_action, obs_2, None)
        measure_time("set act 2", env.set_action, tiger_handle, acts_2)

        # simulate one step
        done = measure_time("step", env.step)

        # get reward
        rewards = measure_time("get reward", env.get_reward, tiger_handle)
        total_reward += sum(rewards)
        measure_time("clear", env.clear_dead)

        step_time = time.time() - start_time
        if i >= skip:
            total_time += step_time
        print("all time: %.2f\n" % (step_time))

        # print info
        print("number of deer: %d"  % env.get_num(deer_handle))
        print("number of tiger: %d" % env.get_num(tiger_handle))
        print("total reward: %d" % total_reward)

        if done:
            print("game over")
            break

    print("FPS", (n_step - skip) / total_time)



from __future__ import absolute_import
from __future__ import division

import math

import pygame
import numpy as np

from magent.renderer.base_renderer import BaseRenderer
from magent.renderer.server import BaseServer


class PyGameRenderer(BaseRenderer):
    def __init__(self):
        super(PyGameRenderer, self).__init__()

    def start(
            self,
            server,
            animation_total=2,
            animation_stop=0,
            resolution=None,
            fps_soft_bound=60,
            background_rgb=(255, 255, 255),
            attack_line_rgb=(0, 0, 0),
            attack_dot_rgb=(0, 0, 0),
            attack_dot_size=0.3,
            text_rgb=(0, 0, 0),
            text_size=16,
            text_spacing=3,
            banner_size=32,
            banner_spacing=3,
            bigscreen_size=72,
            bigscreen_spacing=0,
            grid_rgba=(pygame.Color(0, 0, 0), 30),
            grid_size=7.5,
            grid_min_size=2,
            grid_max_size=100,
            zoom_rate=1 / 30,
            move_rate=4,
            full_screen=False
    ):
        def draw_line(surface, color, a, b):
            pygame.draw.line(
                surface, color,
                (int(round(a[0])), int(round(a[1]))),
                (int(round(b[0])), int(round(b[1])))
            )

        def draw_rect(surface, color, a, w, h):
            pygame.draw.rect(surface, color, pygame.Rect(*map(int, (
                round(a[0]), round(a[1]),
                round(w + a[0] - round(a[0])),
                round(h + a[1] - round(a[1]))))))
            
        def draw_rect_matrix(matrix, color, a, w, h, resolution):
            x, y, w, h = map(int, (round(a[0]), round(a[1]), round(w + a[0] - round(a[0])), round(h + a[1] - round(a[1]))))
            matrix[max(x, 0):min(x + w, resolution[0]), max(y, 0):min(h + y, resolution[1]), :] = color
                    
        def draw_line_matrix(matrix, color, a, b, resolution):
            a = (min(max(0, a[0]), resolution[0] - 1), min(max(0, a[1]), resolution[1] - 1))
            b = (min(max(0, b[0]), resolution[0] - 1), min(max(0, b[1]), resolution[1] - 1))
            a = map(int, (round(a[0]), round(a[1])))
            b = map(int, (round(b[0]), round(b[1])))
            if a[0] == b[0]:
                if a[1] > b[1]:
                    matrix[a[0], b[1]:a[1] + 1] = color
                else:
                    matrix[a[0], a[1]:b[1] + 1] = color
            elif a[1] == b[1]:
                if a[0] > b[0]:
                    matrix[b[0]:a[0] + 1, a[1]] = color
                else:
                    matrix[a[0]:b[0] + 1, a[1]] = color
            else:
                raise NotImplementedError

        if not isinstance(server, BaseServer):
            raise BaseException('property server must be an instance of BaseServer')

        pygame.init()
        pygame.display.init()

        if resolution is None:
            info = pygame.display.Info()
            resolution = info.current_w, info.current_h

        clock = pygame.time.Clock()
        
        if full_screen:
            canvas = pygame.display.set_mode(resolution, pygame.DOUBLEBUF | pygame.FULLSCREEN, 0)
        else:
            canvas = pygame.display.set_mode(resolution, pygame.DOUBLEBUF, 0)

        pygame.display.set_caption('MAgent Renderer Window')
        text_formatter = pygame.font.SysFont(None, text_size, True)
        banner_formatter = pygame.font.SysFont(None, banner_size, True)
        bigscreen_formatter = pygame.font.SysFont(None, bigscreen_size, True)

        map_size, groups, static_info = server.get_info()
        view_position = [map_size[0] / 2 * grid_size - resolution[0] / 2, 
                         map_size[1] / 2 * grid_size - resolution[1] / 2]
        frame_id = 0

        walls  = static_info['wall']

        old_data = None
        new_data = None

        need_static_update = True
        #show_grid = False
        animation_progress = 0
        
        grid_map = np.zeros((resolution[0], resolution[1], 3), dtype=np.int16)

        while True:
            done = False
            status = server.get_status(frame_id)
            triggered = False
            # calculate the relative moues coordinates in the gridworld
            mouse_x, mouse_y = pygame.mouse.get_pos()
            mouse_x = int((mouse_x + view_position[0]) / grid_size)
            mouse_y = int((mouse_y + view_position[1]) / grid_size)
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    pygame.quit()
                    done = True
                elif event.type == pygame.KEYDOWN:
                    #if event.key == pygame.K_g:
                    #    show_grid = not show_grid
                    #else:
                    #    triggered = server.keydown(frame_id, event.key, mouse_x, mouse_y)
                    triggered = server.keydown(frame_id, event.key, mouse_x, mouse_y)
                elif event.type == pygame.MOUSEBUTTONDOWN:
                    if event.button == 4 or event.button == 5:
                        center_before = (
                            (view_position[0] + resolution[0] / 2) / grid_size,
                            (view_position[1] + resolution[1] / 2) / grid_size
                        )
                        if event.button == 5:
                            grid_size = max(grid_size - grid_size * zoom_rate, grid_min_size)
                            need_static_update = True
                        else:
                            grid_size = min(grid_size + grid_size * zoom_rate, grid_max_size)
                            need_static_update = True
                        center_after = (
                            (view_position[0] + resolution[0] / 2) / grid_size,
                            (view_position[1] + resolution[1] / 2) / grid_size
                        )
                        view_position[0] += (center_before[0] - center_after[0]) * grid_size
                        view_position[1] += (center_before[1] - center_after[1]) * grid_size
                    else:
                        triggered = server.mousedown(frame_id, pygame.mouse.get_pressed(), mouse_x, mouse_y)

            pressed = pygame.key.get_pressed()
            if pressed[pygame.K_ESCAPE]:
                pygame.quit()
                done = True

            if pressed[pygame.K_COMMA] or pressed[pygame.K_PERIOD]:
                # center before means the center before zoom operation
                # center after means the center after zoom operation
                # we need to keep that the above two are consistent during zoom operation
                # and hence we need to adjust view_position simultaneously
                center_before = (
                    (view_position[0] + resolution[0] / 2) / grid_size,
                    (view_position[1] + resolution[1] / 2) / grid_size
                )
                if pressed[pygame.K_COMMA]:
                    grid_size = max(grid_size - grid_size * zoom_rate, grid_min_size)
                    need_static_update = True
                else:
                    grid_size = min(grid_size + grid_size * zoom_rate, grid_max_size)
                    need_static_update = True
                center_after = (
                    (view_position[0] + resolution[0] / 2) / grid_size,
                    (view_position[1] + resolution[1] / 2) / grid_size
                )
                view_position[0] += (center_before[0] - center_after[0]) * grid_size
                view_position[1] += (center_before[1] - center_after[1]) * grid_size

            if pressed[pygame.K_LEFT]:
                view_position[0] -= move_rate * grid_size
                need_static_update = True
            if pressed[pygame.K_RIGHT]:
                view_position[0] += move_rate * grid_size
                need_static_update = True
            if pressed[pygame.K_UP]:
                view_position[1] -= move_rate * grid_size
                need_static_update = True
            if pressed[pygame.K_DOWN]:
                view_position[1] += move_rate * grid_size
                need_static_update = True

            if done:
                break

            # x_range: which vertical gridlines should be shown on the display
            # y_range: which horizontal gridlines should be shown on the display
            x_range = (
                max(0, int(math.floor(max(0, view_position[0]) / grid_size))),
                min(map_size[0], int(math.ceil(max(0, view_position[0] + resolution[0]) / grid_size)))
            )

            y_range = (
                max(0, int(math.floor(max(0, view_position[1]) / grid_size))),
                min(map_size[1], int(math.ceil(max(0, view_position[1] + resolution[1]) / grid_size)))
            )

            canvas.fill(background_rgb)

            #if show_grid:
            #    if need_static_update or True:
            #        grids = pygame.Surface(resolution)
            #        grids.set_alpha(grid_rgba[1])
            #        grids.fill(background_rgb)
            #
            #        for i in range(x_range[0], x_range[1] + 1):
            #            draw_line(
            #                canvas, grid_rgba[0],
            #                (i * grid_size - view_position[0], max(0, view_position[1]) - view_position[1]),
            #                (
            #                    i * grid_size - view_position[0],
            #                    min(view_position[1] + resolution[1], map_size[1] * grid_size) - view_position[1]
            #                )
            #            )
            #        for i in range(y_range[0], y_range[1] + 1):
            #            draw_line(
            #                canvas, grid_rgba[0],
            #                (max(0, view_position[0]) - view_position[0], i * grid_size - view_position[1]),
            #                (
            #                    min(view_position[0] + resolution[0], map_size[0] * grid_size) - view_position[0],
            #                    i * grid_size - view_position[1]
            #                )
            #            )

            if new_data is None or animation_progress > animation_total + animation_stop:
                buffered_new_data = server.get_data(
                    frame_id,
                    (view_position[0] / grid_size, (view_position[0] + resolution[0]) / grid_size),
                    (view_position[1] / grid_size, (view_position[1] + resolution[1]) / grid_size)
                )
                if buffered_new_data is None:
                    buffered_new_data = new_data
                old_data = new_data
                new_data = buffered_new_data
                frame_id += 1
                animation_progress = 0

            if new_data is not None:
                if old_data is None and animation_progress == 0:
                    animation_progress = animation_total
                
                if need_static_update:
                    pygame.pixelcopy.surface_to_array(grid_map, canvas)
                    for wall in walls:
                        x, y = wall[0], wall[1]
                        if x >= x_range[0] and x <= x_range[1] and y >= y_range[0] and y <= y_range[1]:
                            draw_rect_matrix(grid_map, (127, 127, 127),
                                             (x * grid_size - view_position[0], y * grid_size - view_position[1]),
                                             grid_size, grid_size, resolution)
                pygame.pixelcopy.array_to_surface(canvas, grid_map)
                
                rate = min(1.0, animation_progress / animation_total)
                for key in new_data[0]:
                    new_prop = new_data[0][key]
                    old_prop = old_data[0][key] if old_data is not None and key in old_data[0] else None
                    new_group = groups[new_prop[2]]
                    old_group = groups[old_prop[2]] if old_prop is not None else None
                    now_prop = [a * (1 - rate) + b * rate for a, b in
                                zip(old_prop, new_prop)] if old_prop is not None else new_prop
                    now_group = [a * (1 - rate) + b * rate for a, b in
                                 zip(old_group, new_group)] if old_group is not None else new_group

                    draw_rect(
                        canvas, (int(now_group[2]), int(now_group[3]), int(now_group[4])),
                        (
                            now_prop[0] * grid_size - view_position[0],
                            now_prop[1] * grid_size - view_position[1]
                        ),
                        now_group[0] * grid_size,
                        now_group[1] * grid_size
                    )

                for key, event_x, event_y in new_data[1]:
                    if not key in new_data[0]:
                        continue
                    new_prop = new_data[0][key]
                    old_prop = old_data[0][key] if old_data is not None and key in old_data[0] else None
                    new_group = groups[new_prop[2]]
                    old_group = groups[old_prop[2]] if old_prop is not None else None
                    now_prop = [a * (1 - rate) + b * rate for a, b in
                                zip(old_prop, new_prop)] if old_prop is not None else new_prop
                    now_group = [a * (1 - rate) + b * rate for a, b in
                                 zip(old_group, new_group)] if old_group is not None else new_group
                    draw_line(
                        canvas, attack_line_rgb,
                        (
                            now_prop[0] * grid_size - view_position[0] + now_group[0] / 2 * grid_size,
                            now_prop[1] * grid_size - view_position[1] + now_group[1] / 2 * grid_size
                        ),
                        (
                            event_x * grid_size - view_position[0] + grid_size / 2,
                            event_y * grid_size - view_position[1] + grid_size / 2
                        )
                    )
                    draw_rect(
                        canvas, attack_dot_rgb,
                        (
                            event_x * grid_size - view_position[0] + grid_size / 2 - attack_dot_size * grid_size / 2,
                            event_y * grid_size - view_position[1] + grid_size / 2 - attack_dot_size * grid_size / 2,
                        ),
                        attack_dot_size * grid_size,
                        attack_dot_size * grid_size
                    )

                if status or triggered or animation_progress < animation_total + animation_stop:
                    animation_progress += 1

                text_fps = text_formatter.render('FPS: {}'.format(int(clock.get_fps())), True, text_rgb)
                text_window = text_formatter.render(
                    'Window: (%.1f, %.1f, %.1f, %.1f)' % (
                        view_position[0], view_position[1],
                        view_position[0] + resolution[0],
                        view_position[1] + resolution[1]
                    ), True, text_rgb
                )

                text_grids = text_formatter.render('Numbers: %d' % len(new_data[0]), True, text_rgb)
                text_mouse = text_formatter.render('Mouse: (%d, %d)' % (mouse_x, mouse_y), True, text_rgb)

                canvas.blit(text_fps, (0, 0))
                canvas.blit(text_window, (0, (text_size + text_spacing) / 1.5))
                canvas.blit(text_grids, (0, (text_size + text_spacing) / 1.5 * 2))
                canvas.blit(text_mouse, (0, (text_size + text_spacing) / 1.5 * 3))

                height_now = 0
                for texts in server.get_banners(frame_id, resolution):
                    content = []
                    width, height = 0, 0
                    for text in texts:
                         text = banner_formatter.render(text[0], True, pygame.Color(*text[1]))
                         content.append((text, width))
                         width += text.get_width()
                         height = max(height, text.get_height())
                    start = (resolution[0] - width) / 2.0
                    for b in content:
                        canvas.blit(b[0], (start + b[1], height_now))
                    height_now += height + banner_spacing

                endscreen_texts = server.get_endscreen(frame_id)
                if endscreen_texts:
                    total_height = 0
                    endscreen_contents = []
                    endscreen = pygame.Surface(resolution)
                    endscreen.set_alpha(230)
                    endscreen.fill(background_rgb)
                    for texts in endscreen_texts:
                        content = []
                        height = 0
                        for text in texts:
                            text = bigscreen_formatter.render(text[0], True, pygame.Color(*text[1]))
                            content.append([text])
                            height = max(height, text.get_height())
                        total_height += height + bigscreen_spacing
                        endscreen_contents.append(content)

                    total_height -= bigscreen_spacing
                    for content in endscreen_contents:
                        height, total_width = 0, 0
                        for b in content:
                            b.append(total_width)
                            total_width += b[0].get_width()
                            height = max(height, b[0].get_height())
                        width_now = (resolution[0] - total_width) / 2.0
                        for b in content:
                            endscreen.blit(b[0], (width_now + b[1], resolution[1] / 2.0 - total_height + height))
                        height_now += height + bigscreen_spacing
                    canvas.blit(endscreen, (0, 0))
            if need_static_update:
                need_static_update = False

            pygame.display.update()
            clock.tick(fps_soft_bound)



"""Deep q network"""

import time

import numpy as np
import tensorflow as tf

from .base import TFBaseModel
from ..common import ReplayBuffer


class DeepQNetwork(TFBaseModel):
    def __init__(self, env, handle, name,
                 batch_size=64, learning_rate=1e-4, reward_decay=0.99,
                 train_freq=1, target_update=2000, memory_size=2 ** 20, eval_obs=None,
                 use_dueling=True, use_double=True, use_conv=True,
                 custom_view_space=None, custom_feature_space=None,
                 num_gpu=1, infer_batch_size=8192, network_type=0):
        """init a model

        Parameters
        ----------
        env: Environment
            environment
        handle: Handle (ctypes.c_int32)
            handle of this group, can be got by env.get_handles
        name: str
            name of this model
        learning_rate: float
        batch_size: int
        reward_decay: float
            reward_decay in TD
        train_freq: int
            mean training times of a sample
        target_update: int
            target will update every target_update batches
        memory_size: int
            weight of entropy loss in total loss
        eval_obs: numpy array
            evaluation set of observation
        use_dueling: bool
            whether use dueling q network
        use_double: bool
            whether use double q network
        use_conv: bool
            use convolution or fully connected layer as state encoder
        num_gpu: int
            number of gpu
        infer_batch_size: int
            batch size while inferring actions
        custom_feature_space: tuple
            customized feature space
        custom_view_space: tuple
            customized feature space
        """
        TFBaseModel.__init__(self, env, handle, name, "tfdqn")
        # ======================== set config  ========================
        self.env = env
        self.handle = handle
        self.view_space = custom_view_space or env.get_view_space(handle)
        self.feature_space = custom_feature_space or env.get_feature_space(handle)
        self.num_actions  = env.get_action_space(handle)[0]

        self.batch_size   = batch_size
        self.learning_rate= learning_rate
        self.train_freq   = train_freq     # train time of every sample (s,a,r,s')
        self.target_update= target_update  # target network update frequency
        self.eval_obs     = eval_obs
        self.infer_batch_size = infer_batch_size  # maximum batch size when infer actions,
        # change this to fit your GPU memory if you meet a OOM

        self.use_dueling  = use_dueling
        self.use_double   = use_double
        self.num_gpu      = num_gpu
        self.network_type = network_type

        self.train_ct = 0

        # ======================= build network =======================
        # input place holder
        self.target = tf.placeholder(tf.float32, [None])
        self.weight = tf.placeholder(tf.float32, [None])

        self.input_view    = tf.placeholder(tf.float32, (None,) + self.view_space)
        self.input_feature = tf.placeholder(tf.float32, (None,) + self.feature_space)
        self.action = tf.placeholder(tf.int32, [None])
        self.mask   = tf.placeholder(tf.float32, [None])
        self.eps = tf.placeholder(tf.float32)  # e-greedy

        # build graph
        with tf.variable_scope(self.name):
            with tf.variable_scope("eval_net_scope"):
                self.eval_scope_name   = tf.get_variable_scope().name
                self.qvalues = self._create_network(self.input_view, self.input_feature, use_conv)

            if self.num_gpu > 1:  # build inference graph for multiple gpus
                self._build_multi_gpu_infer(self.num_gpu)

            with tf.variable_scope("target_net_scope"):
                self.target_scope_name = tf.get_variable_scope().name
                self.target_qvalues = self._create_network(self.input_view, self.input_feature, use_conv)

        # loss
        self.gamma = reward_decay
        self.actions_onehot = tf.one_hot(self.action, self.num_actions)
        td_error = tf.square(self.target - tf.reduce_sum(tf.multiply(self.actions_onehot, self.qvalues), axis=1))
        self.loss = tf.reduce_sum(td_error * self.mask) / tf.reduce_sum(self.mask)

        # train op (clip gradient)
        optimizer = tf.train.AdamOptimizer(learning_rate=learning_rate)
        gradients, variables = zip(*optimizer.compute_gradients(self.loss))
        gradients, _ = tf.clip_by_global_norm(gradients, 5.0)
        self.train_op = optimizer.apply_gradients(zip(gradients, variables))

        # output action
        def out_action(qvalues):
            best_action = tf.argmax(qvalues, axis=1)
            best_action = tf.to_int32(best_action)
            random_action = tf.random_uniform(tf.shape(best_action), 0, self.num_actions, tf.int32)
            should_explore = tf.random_uniform(tf.shape(best_action), 0, 1) < self.eps
            return tf.where(should_explore, random_action, best_action)

        self.output_action = out_action(self.qvalues)
        if self.num_gpu > 1:
            self.infer_out_action = [out_action(qvalue) for qvalue in self.infer_qvalues]

        # target network update op
        self.update_target_op = []
        t_params = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, self.target_scope_name)
        e_params = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, self.eval_scope_name)
        for i in range(len(t_params)):
            self.update_target_op.append(tf.assign(t_params[i], e_params[i]))

        # init tensorflow session
        config = tf.ConfigProto(allow_soft_placement=True, log_device_placement=False)
        config.gpu_options.allow_growth = True
        self.sess = tf.Session(config=config)
        self.sess.run(tf.global_variables_initializer())

        # init replay buffers
        self.replay_buf_len = 0
        self.memory_size = memory_size
        self.replay_buf_view     = ReplayBuffer(shape=(memory_size,) + self.view_space)
        self.replay_buf_feature  = ReplayBuffer(shape=(memory_size,) + self.feature_space)
        self.replay_buf_action   = ReplayBuffer(shape=(memory_size,), dtype=np.int32)
        self.replay_buf_reward   = ReplayBuffer(shape=(memory_size,))
        self.replay_buf_terminal = ReplayBuffer(shape=(memory_size,), dtype=np.bool)
        self.replay_buf_mask     = ReplayBuffer(shape=(memory_size,))
        # if mask[i] == 0, then the item is used for padding, not for training

    def _create_network(self, input_view, input_feature, use_conv=True, reuse=None):
        """define computation graph of network

        Parameters
        ----------
        input_view: tf.tensor
        input_feature: tf.tensor
            the input tensor
        """
        kernel_num  = [32, 32]
        hidden_size = [256]

        if use_conv:  # convolution
            h_conv1 = tf.layers.conv2d(input_view, filters=kernel_num[0], kernel_size=3,
                                       activation=tf.nn.relu, name="conv1", reuse=reuse)
            h_conv2 = tf.layers.conv2d(h_conv1, filters=kernel_num[1], kernel_size=3,
                                       activation=tf.nn.relu, name="conv2", reuse=reuse)
            flatten_view = tf.reshape(h_conv2, [-1, np.prod([v.value for v in h_conv2.shape[1:]])])
            h_view = tf.layers.dense(flatten_view, units=hidden_size[0], activation=tf.nn.relu,
                                     name="dense_view", reuse=reuse)
        else:         # fully connected
            flatten_view = tf.reshape(input_view, [-1, np.prod([v.value for v in input_view.shape[1:]])])
            h_view = tf.layers.dense(flatten_view, units=hidden_size[0], activation=tf.nn.relu)

        h_emb = tf.layers.dense(input_feature,  units=hidden_size[0], activation=tf.nn.relu,
                                name="dense_emb", reuse=reuse)

        dense = tf.concat([h_view, h_emb], axis=1)

        if self.use_dueling:
            value = tf.layers.dense(dense, units=1, name="value", reuse=reuse)
            advantage = tf.layers.dense(dense, units=self.num_actions, use_bias=False,
                                        name="advantage", reuse=reuse)

            qvalues = value + advantage - tf.reduce_mean(advantage, axis=1, keep_dims=True)
        else:
            qvalues = tf.layers.dense(dense, units=self.num_actions, name="value", reuse=reuse)

        return qvalues

    def infer_action(self, raw_obs, ids, policy='e_greedy', eps=0):
        """infer action for a batch of agents

        Parameters
        ----------
        raw_obs: tuple(numpy array, numpy array)
            raw observation of agents tuple(views, features)
        ids: numpy array
            ids of agents
        policy: str
            can be eps-greedy or greedy
        eps: float
            used when policy is eps-greedy

        Returns
        -------
        acts: numpy array of int32
            actions for agents
        """
        view, feature = raw_obs[0], raw_obs[1]

        if policy == 'e_greedy':
            eps = eps
        elif policy == 'greedy':
            eps = 0

        n = len(view)
        batch_size = min(n, self.infer_batch_size)

        if self.num_gpu > 1 and n > batch_size:   # infer by multi gpu in parallel
            ret = self._infer_multi_gpu(view, feature, ids, eps)
        else:                  # infer by splitting big batch in serial
            ret = []
            for i in range(0, n, batch_size):
                beg, end = i, i + batch_size
                ret.append(self.sess.run(self.output_action, feed_dict={
                    self.input_view: view[beg:end],
                    self.input_feature: feature[beg:end],
                    self.eps: eps}))
            ret = np.concatenate(ret)
        return ret

    def _calc_target(self, next_view, next_feature, rewards, terminal):
        """calculate target value"""
        n = len(rewards)
        if self.use_double:
            t_qvalues, qvalues = self.sess.run([self.target_qvalues, self.qvalues],
                                               feed_dict={self.input_view: next_view,
                                                          self.input_feature: next_feature})
            next_value = t_qvalues[np.arange(n), np.argmax(qvalues, axis=1)]
        else:
            t_qvalues = self.sess.run(self.target_qvalues, {self.input_view: next_view,
                                                            self.input_feature: next_feature})
            next_value = np.max(t_qvalues, axis=1)

        target = np.where(terminal, rewards, rewards + self.gamma * next_value)

        return target

    def _add_to_replay_buffer(self, sample_buffer):
        """add samples in sample_buffer to replay buffer"""
        n = 0
        for episode in sample_buffer.episodes():
            v, f, a, r = episode.views, episode.features, episode.actions, episode.rewards

            m = len(r)

            mask = np.ones((m,))
            terminal = np.zeros((m,), dtype=np.bool)
            if episode.terminal:
                terminal[-1] = True
            else:
                mask[-1] = 0

            self.replay_buf_view.put(v)
            self.replay_buf_feature.put(f)
            self.replay_buf_action.put(a)
            self.replay_buf_reward.put(r)
            self.replay_buf_terminal.put(terminal)
            self.replay_buf_mask.put(mask)

            n += m

        self.replay_buf_len = min(self.memory_size, self.replay_buf_len + n)
        return n

    def train(self, sample_buffer, print_every=1000):
        """ add new samples in sample_buffer to replay buffer and train

        Parameters
        ----------
        sample_buffer: magent.utility.EpisodesBuffer
            buffer contains samples
        print_every: int
            print log every print_every batches

        Returns
        -------
        loss: float
            bellman residual loss
        value: float
            estimated state value
        """
        add_num = self._add_to_replay_buffer(sample_buffer)
        batch_size = self.batch_size
        total_loss = 0

        n_batches = int(self.train_freq * add_num / batch_size)
        if n_batches == 0:
            return 0, 0

        print("batch number: %d  add: %d  replay_len: %d/%d" %
              (n_batches, add_num, self.replay_buf_len, self.memory_size))

        start_time = time.time()
        ct = 0
        for i in range(n_batches):
            # fetch a batch
            index = np.random.choice(self.replay_buf_len - 1, batch_size)

            batch_view     = self.replay_buf_view.get(index)
            batch_feature  = self.replay_buf_feature.get(index)
            batch_action   = self.replay_buf_action.get(index)
            batch_reward   = self.replay_buf_reward.get(index)
            batch_terminal = self.replay_buf_terminal.get(index)
            batch_mask     = self.replay_buf_mask.get(index)

            batch_next_view    = self.replay_buf_view.get(index+1)
            batch_next_feature = self.replay_buf_feature.get(index+1)

            batch_target = self._calc_target(batch_next_view, batch_next_feature,
                                             batch_reward, batch_terminal)

            ret = self.sess.run([self.train_op, self.loss], feed_dict={
                self.input_view:    batch_view,
                self.input_feature: batch_feature,
                self.action:        batch_action,
                self.target:        batch_target,
                self.mask:          batch_mask
            })
            loss = ret[1]
            total_loss += loss

            if ct % self.target_update == 0:
                self.sess.run(self.update_target_op)

            if ct % print_every == 0:
                print("batch %5d,  loss %.6f, eval %.6f" % (ct, loss, self._eval(batch_target)))
            ct += 1
            self.train_ct += 1

        total_time = time.time() - start_time
        step_average = total_time / max(1.0, (ct / 1000.0))
        print("batches: %d,  total time: %.2f,  1k average: %.2f" % (ct, total_time, step_average))

        return total_loss / ct if ct != 0 else 0, self._eval(batch_target)

    def _eval(self, target):
        """evaluate estimated q value"""
        if self.eval_obs is None:
            return np.mean(target)
        else:
            return np.mean(self.sess.run([self.qvalues], feed_dict={
                self.input_view: self.eval_obs[0],
                self.input_feature: self.eval_obs[1]
            }))

    def clear_buffer(self):
        """clear replay buffer"""
        self.replay_buf_len = 0
        self.replay_buf_view.clear()
        self.replay_buf_feature.clear()
        self.replay_buf_action.clear()
        self.replay_buf_reward.clear()
        self.replay_buf_terminal.clear()
        self.replay_buf_mask.clear()

    def _build_multi_gpu_infer(self, num_gpu):
        """build inference graph for multi gpus"""
        self.infer_qvalues = []
        self.infer_input_view = []
        self.infer_input_feature = []
        for i in range(num_gpu):
            self.infer_input_view.append(tf.placeholder(tf.float32, (None,) + self.view_space))
            self.infer_input_feature.append(tf.placeholder(tf.float32, (None,) + self.feature_space))
            with tf.variable_scope("eval_net_scope"), tf.device("/gpu:%d" % i):
                self.infer_qvalues.append(self._create_network(self.infer_input_view[i],
                                                               self.infer_input_feature[i], reuse=True))

    def _infer_multi_gpu(self, view, feature, ids, eps):
        """infer action by multi gpu in parallel """
        ret = []
        beg = 0
        while beg < len(view):
            feed_dict = {self.eps: eps}
            for i in range(self.num_gpu):
                end = beg + self.infer_batch_size
                feed_dict[self.infer_input_view[i]] = view[beg:end]
                feed_dict[self.infer_input_feature[i]] = feature[beg:end]
                beg += self.infer_batch_size

            ret.extend(self.sess.run(self.infer_out_action, feed_dict=feed_dict))
        return np.concatenate(ret)



import math
import time

import matplotlib.pyplot as plt
import numpy as np

import magent
from magent.builtin.tf_model import DeepQNetwork
from magent.renderer.server import BaseServer


def load_config(map_size):
    gw = magent.gridworld
    cfg = gw.Config()

    cfg.set({"map_width": map_size, "map_height": map_size})
    cfg.set({"minimap_mode": True})

    cfg.set({"embedding_size": 10})

    small = cfg.register_agent_type(
        "small",
        {'width': 1, 'length': 1, 'hp': 10, 'speed': 2,
         'view_range': gw.CircleRange(6), 'attack_range': gw.CircleRange(1.5),
         'damage': 2, 'step_recover': 0.1,
         'step_reward': -0.001, 'kill_reward': 100, 'dead_penalty': -0.05, 'attack_penalty': -1,
         })

    g0 = cfg.add_group(small)
    g1 = cfg.add_group(small)

    a = gw.AgentSymbol(g0, index='any')
    b = gw.AgentSymbol(g1, index='any')

    cfg.add_reward_rule(gw.Event(a, 'attack', b), receiver=a, value=2)
    cfg.add_reward_rule(gw.Event(b, 'attack', a), receiver=b, value=2)

    return cfg


def generate_map(env, map_size, handles):
    width = map_size
    height = map_size

    init_num = 20

    gap = 3
    leftID, rightID = 0, 1

    # left
    pos = []
    for y in range(10, 45):
        pos.append((width / 2 - 5, y))
        pos.append((width / 2 - 4, y))
    for y in range(50, height // 2 + 25):
        pos.append((width / 2 - 5, y))
        pos.append((width / 2 - 4, y))

    for y in range(height // 2 - 25, height - 50):
        pos.append((width / 2 + 5, y))
        pos.append((width / 2 + 4, y))
    for y in range(height - 45, height - 10):
        pos.append((width / 2 + 5, y))
        pos.append((width / 2 + 4, y))
    env.add_walls(pos=pos, method="custom")

    n = init_num
    side = int(math.sqrt(n)) * 2
    pos = []
    for x in range(width // 2 - gap - side, width // 2 - gap - side + side, 2):
        for y in range((height - side) // 2, (height - side) // 2 + side, 2):
            pos.append([x, y, 0])
    env.add_agents(handles[leftID], method="custom", pos=pos)

    # right
    n = init_num
    side = int(math.sqrt(n)) * 2
    pos = []
    for x in range(width // 2 + gap, width // 2 + gap + side, 2):
        for y in range((height - side) // 2, (height - side) // 2 + side, 2):
            pos.append([x, y, 0])
    env.add_agents(handles[rightID], method="custom", pos=pos)


class BattleServer(BaseServer):
    def __init__(self, path="data/battle_model", total_step=1000, add_counter=10, add_interval=50):
        # some parameter
        map_size = 125
        eps = 0.05

        # init the game
        env = magent.GridWorld(load_config(map_size))

        handles = env.get_handles()
        models = []
        models.append(DeepQNetwork(env, handles[0], 'trusty-battle-game-l', use_conv=True))
        models.append(DeepQNetwork(env, handles[1], 'trusty-battle-game-r', use_conv=True))

        # load model
        models[0].load(path, 0, 'trusty-battle-game-l')
        models[1].load(path, 0, 'trusty-battle-game-r')

        # init environment
        env.reset()
        generate_map(env, map_size, handles)

        # save to member variable
        self.env = env
        self.handles = handles
        self.eps = eps
        self.models = models
        self.map_size = map_size
        self.total_step = total_step
        self.add_interval = add_interval
        self.add_counter = add_counter
        self.done = False
        print(env.get_view2attack(handles[0]))
        plt.show()

    def get_info(self):
        return (self.map_size, self.map_size), self.env._get_groups_info(), {'wall': self.env._get_walls_info()}

    def step(self):
        handles = self.handles
        models = self.models
        env = self.env

        obs = [env.get_observation(handle) for handle in handles]
        ids = [env.get_agent_id(handle) for handle in handles]

        counter = []
        for i in range(len(handles)):
            acts = models[i].infer_action(obs[i], ids[i], 'e_greedy', eps=self.eps)
            env.set_action(handles[i], acts)
            counter.append(np.zeros(shape=env.get_action_space(handles[i])))
            for j in acts:
                counter[-1][j] += 1
        # plt.clf()
        # for c in counter:
        #    plt.bar(range(len(c)), c / np.sum(c))
        # plt.draw()
        # plt.pause(1e-8)

        # code for checking the correctness of observation
        # for channel in range(7):
        #     x = magent.round(list(obs[1][0][0][:,:,channel]), 2)
        #     for row in x:
        #         print row
        #     print("-------------")
        # input()

        done = env.step()
        env.clear_dead()

        return done

    def get_data(self, frame_id, x_range, y_range):
        start = time.time()
        if self.done:
            return None
        self.done = self.step()
        pos, event = self.env._get_render_info(x_range, y_range)
        print(" fps ", 1 / (time.time() - start))
        return pos, event

    def add_agents(self, x, y, g):
        pos = []
        for i in range(-5, 5):
            for j in range(-5, 5):
                pos.append((x + i, y + j))
        self.env.add_agents(self.handles[g], method="custom", pos=pos)

        pos = []
        x = np.random.randint(0, self.map_size - 1)
        y = np.random.randint(0, self.map_size - 1)
        for i in range(-5, 5):
            for j in range(-5, 6):
                pos.append((x + i, y + j))
        self.env.add_agents(self.handles[g ^ 1], method="custom", pos=pos)

    def get_map_size(self):
        return self.map_size, self.map_size

    def get_banners(self, frame_id, resolution):
        red = '{}'.format(self.env.get_num(self.handles[0])), (200, 0, 0)
        vs = ' vs ', (0, 0, 0)
        blue = '{}'.format(self.env.get_num(self.handles[1])), (0, 0, 200)
        result = [(red, vs, blue)]

        tmp = '{} chance(s) remained'.format(
            max(0, self.add_counter)), (0, 0, 0)
        result.append((tmp,))

        tmp = '{} / {} steps'.format(frame_id, self.total_step), (0, 0, 0)
        result.append((tmp,))
        if frame_id % self.add_interval == 0 and frame_id < self.total_step and self.add_counter > 0:
            tmp = 'Please press your left mouse button to add agents', (0, 0, 0)
            result.append((tmp,))
        return result

    def get_status(self, frame_id):
        if frame_id % self.add_interval == 0 and self.add_counter > 0:
            return False
        elif frame_id >= self.total_step or self.done:
            return None
        else:
            return True

    def keydown(self, frame_id, key, mouse_x, mouse_y):
        return False

    def mousedown(self, frame_id, pressed, mouse_x, mouse_y):
        if frame_id % self.add_interval == 0 and frame_id < self.total_step and pressed[0] \
                and self.add_counter > 0 and not self.done:
            self.add_counter -= 1
            pos = []
            for i in range(-5, 5):
                for j in range(-5, 5):
                    pos.append((mouse_x + i, mouse_y + j))
            self.env.add_agents(self.handles[0], method="custom", pos=pos)

            pos = []
            x = np.random.randint(0, self.map_size - 1)
            y = np.random.randint(0, self.map_size - 1)
            for i in range(-5, 6):
                for j in range(-5, 5):
                    pos.append((x + i, y + j))
            self.env.add_agents(self.handles[1], method="custom", pos=pos)
            return True
        return False

    def get_endscreen(self, frame_id):
        if frame_id == self.total_step or self.done:
            if self.env.get_num(self.handles[0]) > self.env.get_num(self.handles[1]):
                return [(("You", (200, 0, 0)), (" win! :)", (0, 0, 0)))]
            else:
                return [(("You", (200, 0, 0)), (" lose. :(", (0, 0, 0)))]
        else:
            return []



"""rename tensorflow models"""

import sys

import magent
from magent.builtin.tf_model import DeepQNetwork

env = magent.GridWorld("battle", map_size=125)

handles = env.get_handles()

rounds = eval(sys.argv[1])

for i in [rounds]:
    model = DeepQNetwork(env, handles[0], "battle")
    print("load %d" % i)
    model.load("data/", i, "selfplay")
    print("save %d" % i)
    model.save("data/battle_model", i)



"""plot curve from many log files"""

import sys
import matplotlib.pyplot as plt
import numpy as np

rec_filename = sys.argv[1]
plot_key = sys.argv[2]
list_col_index = int(sys.argv[3]) if len(sys.argv) > 3 else -1
silent = sys.argv[-1] == '--silent'

def parse_pair(item):
    """parse pair  \tkey: value\t """
    split_index = item.find(":")
    key = item[:split_index].strip()
    value = item[split_index+1:].strip()
    return key, value

def parse_log_file(filename, begin_item_index=0):
    """log_file format   \tkey: value\t key:value\t key:value\t ... """
    ret = {}
    with open(filename, 'r') as fin:
        for line in fin.readlines():
            items = line.split('\t')

            if len(items) < 1:  # ignore error
                continue

            for item in items[begin_item_index:]:
                key, value = parse_pair(item)
                if key not in ret:
                    ret[key] = []
                ret[key].append(value)

    return ret


rec_dict = parse_log_file(rec_filename)


legend = []
data = []
for log_file_name in rec_dict["log_file"]: # parse every file
    log_dict = parse_log_file(log_file_name)
    now = log_dict[plot_key]

    tmp = eval(now[0])
    if isinstance(tmp, list): # is list, expand it
        col_num = len(tmp)
        for row in range(len(now)):
            now[row] = eval(now[row])
        now = np.array(now)

        print(now)

        if list_col_index == -1:
            for col in range(col_num):
                legend.append(log_file_name + "-" + str(col))
                data.append(now[:,col])
        else:
            legend.append(log_file_name)
            data.append(now[:,list_col_index])
    else:  # is a scalar
        for i in range(len(now)):
            now[i] = eval(now[i])
        legend.append(log_file_name)
        data.append(now)

data = np.array(data)

print(legend)
print(data)
plt.plot(data.T)
plt.legend(legend)
plt.savefig(rec_filename + ".png")
if not silent:
    plt.show()



"""gridworld interface"""
from __future__ import absolute_import

import ctypes
import os
import importlib

import numpy as np

from .c_lib import _LIB, as_float_c_array, as_int32_c_array
from .environment import Environment


class GridWorld(Environment):
    # constant
    OBS_INDEX_VIEW = 0
    OBS_INDEX_HP   = 1

    def __init__(self, config, **kwargs):
        """
        Parameters
        ----------
        config: str or Config Object
            if config is a string, then it is a name of builtin config,
                builtin config are stored in python/magent/builtin/config
                kwargs are the arguments to the config
            if config is a Config Object, then parameters are stored in that object
        """
        Environment.__init__(self)

        # if is str, load built in configuration
        if isinstance(config, str):
            # built-in config are stored in python/magent/builtin/config
            try:
                demo_game = importlib.import_module('magent.builtin.config.' + config)
                config = getattr(demo_game, 'get_config')(**kwargs)
            except AttributeError:
                raise BaseException('unknown built-in game "' + config + '"')

        # create new game
        game = ctypes.c_void_p()
        _LIB.env_new_game(ctypes.byref(game), b"GridWorld")
        self.game = game

        # set global configuration
        config_value_type = {
            'map_width': int, 'map_height': int,
            'food_mode': bool, 'turn_mode': bool, 'minimap_mode': bool,
            'revive_mode': bool, 'goal_mode': bool,
            'embedding_size': int,
            'render_dir': str,
        }

        for key in config.config_dict:
            value_type = config_value_type[key]
            if value_type is int:
                _LIB.env_config_game(self.game, key.encode("ascii"), ctypes.byref(ctypes.c_int(config.config_dict[key])))
            elif value_type is bool:
                _LIB.env_config_game(self.game, key.encode("ascii"), ctypes.byref(ctypes.c_bool(config.config_dict[key])))
            elif value_type is float:
                _LIB.env_config_game(self.game, key.encode("ascii"), ctypes.byref(ctypes.c_float(config.config_dict[key])))
            elif value_type is str:
                _LIB.env_config_game(self.game, key.encode("ascii"), ctypes.c_char_p(config.config_dict[key]))

        # register agent types
        for name in config.agent_type_dict:
            type_args = config.agent_type_dict[name]

            # special pre-process for view range and attack range
            for key in [x for x in type_args.keys()]:
                if key == "view_range":
                    val = type_args[key]
                    del type_args[key]
                    type_args["view_radius"] = val.radius
                    type_args["view_angle"]  = val.angle
                elif key == "attack_range":
                    val = type_args[key]
                    del type_args[key]
                    type_args["attack_radius"] = val.radius
                    type_args["attack_angle"]  = val.angle

            length = len(type_args)
            keys = (ctypes.c_char_p * length)(*[key.encode("ascii") for key in type_args.keys()])
            values = (ctypes.c_float * length)(*type_args.values())

            _LIB.gridworld_register_agent_type(self.game, name.encode("ascii"), length, keys, values)

        # serialize event expression, send to C++ engine
        self._serialize_event_exp(config)

        # init group handles
        self.group_handles = []
        for item in config.groups:
            handle = ctypes.c_int32()
            _LIB.gridworld_new_group(self.game, item.encode("ascii"), ctypes.byref(handle))
            self.group_handles.append(handle)

        # init observation buffer (for acceleration)
        self._init_obs_buf()

        # init view space, feature space, action space
        self.view_space = {}
        self.feature_space = {}
        self.action_space = {}
        buf = np.empty((3,), dtype=np.int32)
        for handle in self.group_handles:
            _LIB.env_get_info(self.game, handle, b"view_space",
                              buf.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)))
            self.view_space[handle.value] = (buf[0], buf[1], buf[2])
            _LIB.env_get_info(self.game, handle, b"feature_space",
                                  buf.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)))
            self.feature_space[handle.value] = (buf[0],)
            _LIB.env_get_info(self.game, handle, b"action_space",
                                  buf.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)))
            self.action_space[handle.value] = (buf[0],)

    def reset(self):
        """reset environment"""
        _LIB.env_reset(self.game)

    def add_walls(self, method, **kwargs):
        """add wall to environment

        Parameters
        ----------
        method: str
            can be 'random' or 'custom'
            if method is 'random', then kwargs["n"] is a int
            if method is 'custom', then kwargs["pos"] is a list of coordination

        Examples
        --------
        # add 1000 walls randomly
        >>> env.add_walls(method="random", n=1000)

        # add 3 walls to (1,2), (4,5) and (9, 8) in map
        >>> env.add_walls(method="custom", pos=[(1,2), (4,5), (9,8)])
        """
        # handle = -1 for walls
        kwargs["dir"] = 0
        self.add_agents(-1, method, **kwargs)

    # ====== AGENT ======
    def new_group(self, name):
        """register a new group into environment"""
        handle = ctypes.c_int32()
        _LIB.gridworld_new_group(self.game, ctypes.c_char_p(name.encode("ascii")), ctypes.byref(handle))
        return handle

    def add_agents(self, handle, method, **kwargs):
        """add agents to environment

        Parameters
        ----------
        handle: group handle
        method: str
            can be 'random' or 'custom'
            if method is 'random', then kwargs["n"] is a int
            if method is 'custom', then kwargs["pos"] is a list of coordination

        Examples
        --------
        # add 1000 walls randomly
        >>> env.add_agents(handle, method="random", n=1000)

        # add 3 agents to (1,2), (4,5) and (9, 8) in map
        >>> env.add_agents(handle, method="custom", pos=[(1,2), (4,5), (9,8)])
        """
        if method == "random":
            _LIB.gridworld_add_agents(self.game, handle, int(kwargs["n"]), b"random", 0, 0, 0)
        elif method == "custom":
            n = len(kwargs["pos"])
            pos = np.array(kwargs["pos"], dtype=np.int32)
            if len(pos) <= 0:
                return
            if pos.shape[1] == 3:  # if has dir
                xs, ys, dirs = pos[:, 0], pos[:, 1], pos[:, 2]
            else:                  # if do not has dir, use zero padding
                xs, ys, dirs = pos[:, 0], pos[:, 1], np.zeros((n,), dtype=np.int32)
            # copy again, to make these arrays continuous in memory
            xs, ys, dirs = np.array(xs), np.array(ys), np.array(dirs)
            _LIB.gridworld_add_agents(self.game, handle, n, b"custom", as_int32_c_array(xs),
                                      as_int32_c_array(ys), as_int32_c_array(dirs))
        elif method == "fill":
            x, y = kwargs["pos"][0], kwargs["pos"][1]
            width, height = kwargs["size"][0], kwargs["size"][1]
            dir = kwargs.get("dir", np.zeros_like(x))
            bind = np.array([x, y, width, height, dir], dtype=np.int32)
            _LIB.gridworld_add_agents(self.game, handle, 0,  b"fill", as_int32_c_array(bind),
                                      0, 0, 0)
        elif method == "maze":
            # TODO: implement maze add
            x_start, y_start, x_end, y_end = kwargs["pos"][0], kwargs["pos"][1], kwargs["pos"][2], kwargs["pos"][3]
            thick = kwargs["pos"][4]
            bind = np.array([x_start, y_start, x_end, y_end, thick], dtype=np.int32)
            _LIB.gridworld_add_agents(self.game, handle, 0, b"maze", as_int32_c_array(bind),
                                      0, 0, 0)
        else:
            print("Unknown type of position")
            exit(-1)

    # ====== RUN ======
    def _get_obs_buf(self, group, key, shape, dtype):
        """get buffer to receive observation from c++ engine"""
        obs_buf = self.obs_bufs[key]
        if group in obs_buf:
            ret = obs_buf[group]
            if shape != ret.shape:
                ret.resize(shape, refcheck=False)
        else:
            ret = obs_buf[group] = np.empty(shape=shape, dtype=dtype)

        return ret

    def _init_obs_buf(self):
        """init observation buffer"""
        self.obs_bufs = []
        self.obs_bufs.append({})
        self.obs_bufs.append({})

    def get_observation(self, handle):
        """ get observation of a whole group

        Parameters
        ----------
        handle : group handle

        Returns
        -------
        obs : tuple (views, features)
            views is a numpy array, whose shape is n * view_width * view_height * n_channel
            features is a numpy array, whose shape is n * feature_size
            for agent i, (views[i], features[i]) is its observation at this step
        """
        view_space = self.view_space[handle.value]
        feature_space = self.feature_space[handle.value]
        no = handle.value

        n = self.get_num(handle)
        view_buf = self._get_obs_buf(no, self.OBS_INDEX_VIEW, (n,) + view_space, np.float32)
        feature_buf = self._get_obs_buf(no, self.OBS_INDEX_HP, (n,) + feature_space, np.float32)

        bufs = (ctypes.POINTER(ctypes.c_float) * 2)()
        bufs[0] = as_float_c_array(view_buf)
        bufs[1] = as_float_c_array(feature_buf)
        _LIB.env_get_observation(self.game, handle, bufs)

        return view_buf, feature_buf

    def set_action(self, handle, actions):
        """ set actions for whole group

        Parameters
        ----------
        handle: group handle
        actions: numpy array
            the dtype of actions must be int32
        """
        assert isinstance(actions, np.ndarray)
        assert actions.dtype == np.int32
        _LIB.env_set_action(self.game, handle, actions.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)))

    def step(self):
        """simulation one step after set actions

        Returns
        -------
        done: bool
            whether the game is done
        """
        done = ctypes.c_int32()
        _LIB.env_step(self.game, ctypes.byref(done))
        return bool(done)

    def get_reward(self, handle):
        """ get reward for a whole group

        Returns
        -------
        rewards: numpy array (float32)
            reward for all the agents in the group
        """
        n = self.get_num(handle)
        buf = np.empty((n,), dtype=np.float32)
        _LIB.env_get_reward(self.game, handle,
                            buf.ctypes.data_as(ctypes.POINTER(ctypes.c_float)))
        return buf

    def clear_dead(self):
        """ clear dead agents in the engine
        must be called after step()
        """
        _LIB.gridworld_clear_dead(self.game)

    # ====== INFO ======
    def get_handles(self):
        """ get all group handles in the environment """
        return self.group_handles

    def get_num(self, handle):
        """ get the number of agents in a group"""
        num = ctypes.c_int32()
        _LIB.env_get_info(self.game, handle, b'num', ctypes.byref(num))
        return num.value

    def get_action_space(self, handle):
        """get action space

        Returns
        -------
        action_space : tuple
        """
        return self.action_space[handle.value]

    def get_view_space(self, handle):
        """get view space

        Returns
        -------
        view_space : tuple
        """
        return self.view_space[handle.value]

    def get_feature_space(self, handle):
        """ get feature space

        Returns
        -------
        feature_space : tuple
        """
        return self.feature_space[handle.value]

    def get_agent_id(self, handle):
        """ get agent id

        Returns
        -------
        ids : numpy array (int32)
            id of all the agents in the group
        """
        n = self.get_num(handle)
        buf = np.empty((n,), dtype=np.int32)
        _LIB.env_get_info(self.game, handle, b"id",
                          buf.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)))
        return buf

    def get_alive(self, handle):
        """ get alive status of agents in a group

        Returns
        -------
        alives: numpy array (bool)
            whether the agents are alive
        """
        n = self.get_num(handle)
        buf = np.empty((n,), dtype=np.bool)
        _LIB.env_get_info(self.game, handle, b"alive",
                          buf.ctypes.data_as(ctypes.POINTER(ctypes.c_bool)))
        return buf

    def get_pos(self, handle):
        """ get position of agents in a group

        Returns
        -------
        pos: numpy array (int)
            the shape of pos is (n, 2)
        """
        n = self.get_num(handle)
        buf = np.empty((n, 2), dtype=np.int32)
        _LIB.env_get_info(self.game, handle, b"pos",
                          buf.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)))
        return buf

    def get_view2attack(self, handle):
        """ get a matrix with the same size of view_range,
            if element >= 0, then it means it is a attackable point, and the corresponding
                                    action number is the value of that element
        Returns
        -------
        attack_back: int
        buf: numpy array
            map attack action into view
        """
        size = self.get_view_space(handle)[0:2]
        buf = np.empty(size, dtype=np.int32)
        attack_base = ctypes.c_int32()
        _LIB.env_get_info(self.game, handle, b"view2attack",
                          buf.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)))
        _LIB.env_get_info(self.game, handle, b"attack_base",
                          ctypes.byref(attack_base))
        return attack_base.value, buf

    def get_global_minimap(self, height, width):
        """ compress global map into a minimap of given size
        Parameters
        ----------
        height: int
            the height of minimap
        width:  int
            the width of minimap

        Returns
        -------
        minimap : numpy array
            the shape (n_group + 1, height, width)
        """
        buf = np.empty((height, width, len(self.group_handles)), dtype=np.float32)
        buf[0, 0, 0] = height
        buf[0, 0, 1] = width
        _LIB.env_get_info(self.game, -1, b"global_minimap",
                          buf.ctypes.data_as(ctypes.POINTER(ctypes.c_float)))
        return buf

    def set_seed(self, seed):
        """ set random seed of the engine"""
        _LIB.env_config_game(self.game, b"seed", ctypes.byref(ctypes.c_int(seed)))

    # ====== RENDER ======
    def set_render_dir(self, name):
        """ set directory to save render file"""
        if not os.path.exists(name):
            os.mkdir(name)
        _LIB.env_config_game(self.game, b"render_dir", name.encode("ascii"))

    def render(self):
        """ render a step """
        _LIB.env_render(self.game)

    def _get_groups_info(self):
        """ private method, for interactive application"""
        n = len(self.group_handles)
        buf = np.empty((n, 5), dtype=np.int32)
        _LIB.env_get_info(self.game, -1, b"groups_info",
                          buf.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)))
        return buf

    def _get_walls_info(self):
        """ private method, for interactive application"""
        n = 100 * 100
        buf = np.empty((n, 2), dtype=np.int32)
        _LIB.env_get_info(self.game, -1, b"walls_info",
                          buf.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)))
        n = buf[0, 0]  # the first line is the number of walls
        return buf[1:1+n]

    def _get_render_info(self, x_range, y_range):
        """ private method, for interactive application"""
        n = 0
        for handle in self.group_handles:
            n += self.get_num(handle)

        buf = np.empty((n+1, 4), dtype=np.int32)
        buf[0] = x_range[0], y_range[0], x_range[1], y_range[1]
        _LIB.env_get_info(self.game, -1, b"render_window_info",
                          buf.ctypes.data_as(ctypes.POINTER((ctypes.c_int32))))

        # the first line is for the number of agents in the window range
        info_line = buf[0]
        agent_ct, attack_event_ct = info_line[0], info_line[1]
        buf = buf[1:1 + info_line[0]]

        agent_info = {}
        for item in buf:
            agent_info[item[0]] = [item[1], item[2], item[3]]

        buf = np.empty((attack_event_ct, 3), dtype=np.int32)
        _LIB.env_get_info(self.game, -1, b"attack_event",
                          buf.ctypes.data_as(ctypes.POINTER((ctypes.c_int32))))
        attack_event = buf

        return agent_info, attack_event

    def __del__(self):
        _LIB.env_delete_game(self.game)

    # ====== PRIVATE ======
    def _serialize_event_exp(self, config):
        """serialize event expression and sent them to game engine"""
        game = self.game

        # collect agent symbol
        symbol2int = {}
        config.symbol_ct = 0

        def collect_agent_symbol(node, config):
            for item in node.inputs:
                if isinstance(item, EventNode):
                    collect_agent_symbol(item, config)
                elif isinstance(item, AgentSymbol):
                    if item not in symbol2int:
                        symbol2int[item] = config.symbol_ct
                        config.symbol_ct += 1

        for rule in config.reward_rules:
            on = rule[0]
            receiver = rule[1]
            for symbol in receiver:
                if symbol not in symbol2int:
                    symbol2int[symbol] = config.symbol_ct
                    config.symbol_ct += 1
            collect_agent_symbol(on, config)

        # collect event node
        event2int = {}
        config.node_ct = 0

        def collect_event_node(node, config):
            if node not in event2int:
                event2int[node] = config.node_ct
                config.node_ct += 1
            for item in node.inputs:
                if isinstance(item, EventNode):
                    collect_event_node(item, config)

        for rule in config.reward_rules:
            collect_event_node(rule[0], config)

        # send to C++ engine
        for sym in symbol2int:
            no = symbol2int[sym]
            _LIB.gridworld_define_agent_symbol(game, no, sym.group, sym.index)

        for event in event2int:
            no = event2int[event]
            inputs = np.zeros_like(event.inputs, dtype=np.int32)
            for i, item in enumerate(event.inputs):
                if isinstance(item, EventNode):
                    inputs[i] = event2int[item]
                elif isinstance(item, AgentSymbol):
                    inputs[i] = symbol2int[item]
                else:
                    inputs[i] = item
            n_inputs = len(inputs)
            _LIB.gridworld_define_event_node(game, no, event.op, as_int32_c_array(inputs), n_inputs)

        for rule in config.reward_rules:
            # rule = [on, receiver, value, terminal]
            on = event2int[rule[0]]

            receiver = np.zeros_like(rule[1], dtype=np.int32)
            for i, item in enumerate(rule[1]):
                receiver[i] = symbol2int[item]
            if len(rule[2]) == 1 and rule[2][0] == 'auto':
                value = np.zeros(receiver, dtype=np.float32)
            else:
                value = np.array(rule[2], dtype=np.float32)
            n_receiver = len(receiver)
            _LIB.gridworld_add_reward_rule(game, on, as_int32_c_array(receiver),
                                           as_float_c_array(value), n_receiver, rule[3])


'''
the following classes are for reward description
'''
class EventNode:
    """an AST node of the event expression"""
    OP_AND = 0
    OP_OR  = 1
    OP_NOT = 2

    OP_KILL = 3
    OP_AT   = 4
    OP_IN   = 5
    OP_COLLIDE = 6
    OP_ATTACK  = 7
    OP_DIE  = 8
    OP_IN_A_LINE = 9
    OP_ALIGN = 10

    # can extend more operation below

    def __init__(self):
        # for non-leaf node
        self.op = None
        # for leaf node
        self.predicate = None

        self.inputs = []

    def __call__(self, subject, predicate, *args):
        node = EventNode()
        node.predicate = predicate
        if predicate == 'kill':
            node.op = EventNode.OP_KILL
            node.inputs = [subject, args[0]]
        elif predicate == 'at':
            node.op = EventNode.OP_AT
            coor = args[0]
            node.inputs = [subject, coor[0], coor[1]]
        elif predicate == 'in':
            node.op = EventNode.OP_IN
            coor = args[0]
            x1, y1 = min(coor[0][0], coor[1][0]), min(coor[0][1], coor[1][1])
            x2, y2 = max(coor[0][0], coor[1][0]), max(coor[0][1], coor[1][1])
            node.inputs = [subject, x1, y1, x2, y2]
        elif predicate == 'attack':
            node.op = EventNode.OP_ATTACK
            node.inputs = [subject, args[0]]
        elif predicate == 'kill':
            node.op = EventNode.OP_KILL
            node.inputs = [subject, args[0]]
        elif predicate == 'collide':
            node.op = EventNode.OP_COLLIDE
            node.inputs = [subject, args[0]]
        elif predicate == 'die':
            node.op = EventNode.OP_DIE
            node.inputs = [subject]
        elif predicate == 'in_a_line':
            node.op = EventNode.OP_IN_A_LINE
            node.inputs = [subject]
        elif predicate == 'align':
            node.op = EventNode.OP_ALIGN
            node.inputs = [subject]
        else:
            raise Exception("invalid predicate of event " + predicate)
        return node

    def __and__(self, other):
        node = EventNode()
        node.op = EventNode.OP_AND
        node.inputs = [self, other]
        return node

    def __or__(self, other):
        node = EventNode()
        node.op = EventNode.OP_OR
        node.inputs = [self, other]
        return node

    def __invert__(self):
        node = EventNode()
        node.op = EventNode.OP_NOT
        node.inputs = [self]
        return node

Event = EventNode()

class AgentSymbol:
    """symbol to represent some agents"""
    def __init__(self, group, index):
        """ define a agent symbol, it can be the object or subject of EventNode

        group: group handle
            it is the return value of cfg.add_group()
        index: int or str
            int: a deterministic integer id
            str: can be 'all' or 'any', represents all or any agents in a group
        """
        self.group = group if group is not None else -1
        if index == 'any':
            self.index = -1
        elif index == 'all':
            self.index = -2
        else:
            assert isinstance(self.index, int), "index must be a deterministic int"
            self.index = index

    def __str__(self):
        return 'agent(%d,%d)' % (self.group, self.index)


class Config:
    """configuration class of gridworld game"""
    def __init__(self):
        self.config_dict = {}
        self.agent_type_dict = {}
        self.groups = []
        self.reward_rules = []

    def set(self, args):
        """ set parameters of global configuration

        Parameters
        ----------
        args : dict
            key value pair of the configuration
        """
        for key in args:
            self.config_dict[key] = args[key]

    def register_agent_type(self, name, attr):
        """ register an agent type

        Parameters
        ----------
        name : str
            name of the type (should be unique)
        attr: dict
            key value pair of the agent type
            see notes below to know the available attributes

        Notes
        -----
        height: int, height of agent body
        width:  int, width of agent body
        speed:  float, maximum speed, i.e. the radius of move circle of the agent
        hp:     float, maximum health point of the agent
        view_range: gw.CircleRange or gw.SectorRange

        damage: float, attack damage
        step_recover: float, step recover of health point (can be negative)
        kill_supply: float, the hp gain when kill this type of agents

        step_reward: float, reward get in every step
        kill_reward: float, reward gain when kill this type of agent
        dead_penalty: float, reward get when dead
        attack_penalty: float, reward get when perform an attack (this is used to make agents do not attack blank grid)
        """
        if name in self.agent_type_dict:
            raise Exception("type name %s already exists" % name)
        self.agent_type_dict[name] = attr
        return name

    def add_group(self, agent_type):
        """ add a group to the configuration

        Returns
        -------
        group_handle : int
            a handle for the new added group
        """
        no = len(self.groups)
        self.groups.append(agent_type)
        return no

    def add_reward_rule(self, on, receiver, value, terminal=False):
        """ add a reward rule

        Some note:
        1. if the receiver is not a deterministic agent,
           it must be one of the agents involved in the triggering event

        Parameters
        ----------
        on: Expr
            a bool expression of the trigger event
        receiver:  (list of) AgentSymbol
            receiver of this reward rule
        value: (list of) float
            value to assign
        terminal: bool
            whether this event will terminate the game
        """
        if not (isinstance(receiver, tuple) or isinstance(receiver, list)):
            assert not (isinstance(value, tuple) or isinstance(value, tuple))
            receiver = [receiver]
            value = [value]
        if len(receiver) != len(value):
            raise Exception("the length of receiver and value should be equal")
        self.reward_rules.append([on, receiver, value, terminal])


class CircleRange:
    def __init__(self, radius):
        """ define a circle range for attack or view

        Parameters
        ----------
        radius : float
        """
        self.radius = radius
        self.angle  = 360

    def __str__(self):
        return 'circle(%g)' % self.radius


class SectorRange:
    def __init__(self, radius, angle):
        """ define a sector range for attack or view

        Parameters
        ----------
        radius : float
        angle :  float
            angle should be less than 180
        """
        self.radius = radius
        self.angle  = angle
        if self.angle >= 180:
            raise Exception("the angle of a sector should be smaller than 180 degree")

    def __str__(self):
        return 'sector(%g, %g)' % (self.radius, self.angle)
